"""Actions: instructions for the MTA, based on policy module output"""
import functools
from chapps.config import config
from chapps.policy import GreylistingPolicy


class PostfixActions:
    """Superclass for Postfix action adapters"""

    @staticmethod
    def dunno(*args, **kwargs):
        return "DUNNO"

    @staticmethod
    def okay(*args, **kwargs):
        return "OK"

    ok = okay

    @staticmethod
    def defer_if_permit(msg, *args, **kwargs):
        return f"DEFER_IF_PERMIT {msg}"

    @staticmethod
    def reject(msg, *args, **kwargs):
        return f"REJECT {msg}"

    @staticmethod
    def prepend(msg, *args, **kwargs):
        new_header = kwargs.get("prepend", None)
        if new_header is None or len(new_header) < 5:
            raise ValueError(
                f"Prepended header expected to be at least 5 chars in length."
            )
        return f"PREPEND {new_header}"

    def __init__(self, cfg=None, extended_status=True):
        self.config = cfg or config
        self.extended_status = (
            extended_status
        )  # intended to allow choice whether to use extended status; TODO: not currently implemented

    def _get_closure_for(self, decision):
        """Setup the prescribed closure for generating SMTP action directives"""
        action_config = getattr(self.config, decision, None)
        if not action_config:
            raise ValueError(
                f"Action config for {self.__class__.__name__} does not contain a key named {decision}"
            )
        action_tokens = action_config.split(" ")
        action = action_tokens[0]
        try:
            i = int(action)  #  if the first token is a number, its a directive
        except ValueError:  #  first token was a string, and therefore refers to a method
            af = getattr(self, action, None)
            if af:
                return af
            action_func = getattr(PostfixActions, action, None)
            if not action_func:
                action_func = getattr(self.__class__, action, None)
            if not action_func:
                raise NotImplementedError(
                    f"Action {action} is not implemented by PostfixActions or by {self.__class__.__name__}"
                )
        else:
            action_func = lambda reason, ppr, *args, **kwargs: action_config.format(
                reason=reason
            )
        setattr(self, action, action_func)
        return action_func

    def _get_message_for(self, decision, config_name=None):
        """Grab a status message for a decision from the config, optionally with another name"""
        msg_key = config_name or decision
        msg = getattr(self, msg_key, None)
        if not msg:
            raise ValueError(
                f"There is no key {msg_key} in the config for {self.__class__.__name__} or its policy"
            )
        return msg

    def _mangle_action(self, action):
        """Policy decisions which are also reserved words will need to be altered"""
        if action == "pass":
            return "passing"
        return action

    def action_for(self, *args, **kwargs):
        raise NotImplementedError(
            f"Subclasses of {self.__class__.__name__} must define the method action_for() for themselves, to map policy module response (decision) strings onto Postfix action directives."
        )


class PostfixPassfailActions(PostfixActions):
    def __init__(self, cfg=None):
        super().__init__(cfg)

    def _get_closure_for(self, decision, msg_key=None):
        """Create a closure for formatting these messages and store it on self.<decision>, and also return it"""
        msg_key = msg_key or decision
        msg = getattr(self.config, msg_key, None)
        if not msg:
            raise ValueError(
                f"The key {msg_key} is not defined in the config for {self.__class__.__name__} or its policy"
            )
        msg_tokens = msg.split(" ")
        msg_text = ""
        if msg_tokens[0] == "OK":
            func = PostfixActions.okay
        elif msg_tokens[0] == "DUNNO":
            func = PostfixActions.dunno
        elif msg_tokens[0] == "DEFER_IF_PERMIT":
            func = PostfixActions.defer_if_permit
            msg_text = " ".join(msg_tokens[1:])
        elif msg_tokens[0] == "REJECT" or msg_tokens[0] == "554":
            func = PostfixActions.reject
            msg_text = " ".join(msg_tokens[1:])
        else:
            raise NotImplementedError(
                f"Pass-fail closure creation for Postfix directive {msg_tokens[0]} is not yet available."
            )
        action = self.__prepend_action_with_message(func, msg_text)
        setattr(self, decision, action)
        return action

    def __prepend_action_with_message(self, func, prepend_msg_text):
        ### avoiding use of nonlocal required if definition is embedded inline in calling procedure
        def action(message="", *args, **kwargs):
            msg_text = prepend_msg_text
            if len(message) > 0:
                msg_text = " ".join([msg_text, message])
            return func(msg_text, *args, **kwargs)

        return action

    def action_for(self, pf_result):
        if pf_result:  # True / pass
            action_name = "passing"
        else:  # False / fail
            action_name = "fail"
        return getattr(self, action_name, None)

    def __getattr__(self, attrname, *args, **kwargs):
        attrname = self._mangle_action(attrname)
        if attrname == "passing":
            msg_key = "acceptance_message"
        elif attrname == "fail":
            msg_key = "rejection_message"
        else:
            raise NotImplementedError(f"Pass-fail actions do not include {attrname}")
        return self._get_closure_for(attrname, msg_key)


class PostfixOQPActions(PostfixPassfailActions):
    def __init__(self, cfg=None):
        super().__init__(cfg)
        self.config = self.config.policy_oqp


class PostfixGRLActions(PostfixPassfailActions):
    def __init__(self, cfg=None):
        super().__init__(cfg)
        self.config = self.config.policy_grl


class PostfixSPFActions(PostfixActions):
    greylisting_policy = GreylistingPolicy()

    @staticmethod
    def greylist(msg, *args, **kwargs):
        ppr = kwargs.get("ppr", None)
        if ppr is None:
            raise ValueError(
                f"PostfixSPFActions.greylist() expects a ppr= kwarg providing the PPR for greylisting."
            )
        if PostfixSPFActions.greylisting_policy.approve_policy_request(ppr):
            passing = PostfixSPFActions().action_for("pass")
            return passing(msg, ppr, *args, **kwargs)
        if len(msg) == 0:
            msg = "due to SPF enforcement policy"
        return PostfixGRLActions().fail(msg, ppr, *args, **kwargs)

    def __init__(self, cfg=None):
        super().__init__(cfg)
        self.config = self.config.actions_spf

    def _mangle_action(self, action):
        if action == "none" or action == "neutral":
            action = "none_neutral"
        else:
            action = super()._mangle_action(action)
        return action

    def action_for(self, spf_result):
        spf_result = self._mangle_action(spf_result)
        action = getattr(self, spf_result, None)
        if action:
            return action
        return self._get_closure_for(spf_result)  #  this memoizes its result



"""routes for live access to CHAPPS state"""
from typing import List, Optional
from fastapi import status, APIRouter, Body, HTTPException
from sqlalchemy.orm import sessionmaker
from .users import user_quota_assoc, user_domains_assoc
from .common import load_model_with_assoc
from ..dbsession import sql_engine
from ..models import User, Domain, LiveQuotaResp, TextResp, DomainUserMapResp
from ...policy import OutboundQuotaPolicy, SenderDomainAuthPolicy
from ...config import config
import hashlib
import logging

logger = logging.getLogger(__name__)
Session = sessionmaker(sql_engine)

api = APIRouter(
    prefix="/live",
    tags=["live"],
    responses={
        status.HTTP_404_NOT_FOUND: {"description": "Resource not found."},
        status.HTTP_400_BAD_REQUEST: {
            "description": (
                "One of either user_id or name must be provided. "
                "If both are present, user_id is preferred"
            )
        },
    },
)

# define some useful functions
load_user_with_quota = load_model_with_assoc(User, assoc=[user_quota_assoc])


@api.get("/quota/{user_id}", response_model=LiveQuotaResp)
async def get_current_quota_remaining_for_user(
    user_id: int = 0, name: Optional[str] = Body(None)
):
    """
    Accepts the id of the user whose remaining quota should be checked.<br/>
    Returns the instantaneous number of available send attempts in
    `response`
    """
    user, assoc_d, remarks = load_user_with_quota(user_id, name)
    quota = assoc_d[user_quota_assoc.assoc_name]
    oqp = OutboundQuotaPolicy()
    response, more_remarks = oqp.current_quota(user.name, quota)
    return LiveQuotaResp.send(response, remarks=remarks + more_remarks)


@api.delete("/quota/{user_id}", response_model=LiveQuotaResp)
async def reset_live_quota_for_user(
    user_id: int = 0, name: Optional[str] = Body(None)
):
    """
    Accepts the id of the user whose quota should be reset.<br/>
    Returns the number of send attempts dropped in `response`
    """
    user, assoc_d, remarks = load_user_with_quota(user_id, name)
    quota = assoc_d[user_quota_assoc.assoc_name]
    oqp = OutboundQuotaPolicy()
    response, more_remarks = oqp.reset_quota(user.name)
    if not quota:
        remarks.append(f"User {user.name} has no assigned quota.")
    logger.info(" ".join(more_remarks))  # log only reset message
    return LiveQuotaResp.send(response, remarks=remarks + more_remarks)


@api.post("/quota/{user_id}", response_model=LiveQuotaResp)
async def refresh_quota_policy_for_user(
    user_id: int = 0, name: Optional[str] = Body(None)
):
    """
    Accepts the id of the user whose quota policy should be refreshed.<br/>
    Returns the new remaining quota after the policy update in `response`
    """
    user, assoc_d, remarks = load_user_with_quota(user_id, name)
    quota = assoc_d[user_quota_assoc.assoc_name]
    oqp = OutboundQuotaPolicy()
    if not quota:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Cannot load quota for user {user.name}: none assigned.",
        )
    remarks.append(f"Quota policy config cache reset for {user.name}")
    response, more_remarks = oqp.refresh_policy_cache(user.name, quota)
    return LiveQuotaResp.send(response, remarks=remarks + more_remarks)


@api.post(
    "/config/write/",
    response_model=TextResp,
    responses={
        status.HTTP_401_UNAUTHORIZED: {
            "description": "Password does not match."
        }
    },
)
async def refresh_config_on_disk(passcode: str = Body(...)):
    """
    Writes the current effective config to disk.<br/>
    Requires the CHAPPS password to be provided as the body.<br/>
    If line transmission security is an issue, an SSL proxy layer will
      be required.  This is true for the entire application.
    """
    if (
        hashlib.sha256(
            passcode.encode(config.chapps.payload_encoding)
        ).hexdigest()
        == config.chapps.password
    ):
        response = config.write()
        return TextResp.send(str(response))
    raise HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Password does not match.",
    )


@api.get("/sda/", response_model=DomainUserMapResp)
async def sda_batch_peek(domain_ids: List[int], user_ids: List[int]):
    """
    Accepts `domain_ids` and `user_ids` as body arguments: lists of
    integer object ids.<br/>
    Looks at current authorizations for all domain-user combinations.<br/>
    Returns their cache status as a dict of dicts, keyed as:
    `sda[domain][user] = SDAStatus`
    """
    sda = SenderDomainAuthPolicy()
    with Session() as sess:
        domain_names = list(
            sess.scalars(Domain.select_names_by_id(domain_ids))
        )
        user_names = list(sess.scalars(User.select_names_by_id(user_ids)))
    return DomainUserMapResp.send(
        sda.bulk_check_policy_cache(user_names, domain_names)
    )


@api.get("/sda/{domain_name}/for/{user_name}", response_model=TextResp)
async def sda_peek(domain_name: str, user_name: str):
    """
    Accepts url-encoded domain name and user name as path arguments.<br/>
    Returns status of cached SDA for the named user and domain,
    i.e. is this user allowed to transmit email apparently from this domain
    """
    sda = SenderDomainAuthPolicy()
    return TextResp.send(sda.check_policy_cache(user_name, domain_name))


@api.delete("/sda/", response_model=TextResp)
async def sda_batch_clear(domain_ids: List[int], user_ids: List[int]):
    """
    Clears all domain - user mappings by iterating through both lists.
    """
    sda = SenderDomainAuthPolicy()
    with Session() as sess:
        domain_names = list(
            sess.scalars(Domain.select_names_by_id(domain_ids))
        )
        user_names = list(sess.scalars(User.select_names_by_id(user_ids)))
    sda.bulk_clear_policy_cache(user_names, domain_names)
    return TextResp.send("SDA cache cleared for specified domains x users.")


@api.delete("/sda/{domain_name}/for/{user_name}", response_model=TextResp)
async def sda_clear(domain_name: str, user_name: str):
    """
    Accepts url-encoded domain name and user name of SDA to clear.</br>
    Returns the status of the SDA prior to clearing.
    """
    sda = SenderDomainAuthPolicy()
    return TextResp.send(sda.clear_policy_cache(user_name, domain_name))



"""Policy-configuration source data adapters"""
import mariadb

# import re ### needed if implementing domain_re_for_user()
import logging
from chapps.config import config
from contextlib import contextmanager

logger = logging.getLogger(__name__)  # pragma: no cover


class PolicyConfigAdapter:
    """Base class for DB adapters for policy config parameter access"""

    user_table = (
        "CREATE TABLE IF NOT EXISTS users ("  # pragma: no cover
        "id BIGINT AUTO_INCREMENT PRIMARY KEY,"
        "name VARCHAR(128) UNIQUE NOT NULL"
        ")"
    )
    userid_query = "SELECT id FROM users WHERE name = %(user)s"

    def __init__(
        self,
        *,
        db_host=None,
        db_port=None,
        db_name=None,
        db_user=None,
        db_pass=None,
        autocommit=True,
    ):
        self.host = db_host or config.adapter.db_host or "127.0.0.1"
        self.port = db_port or config.adapter.db_port or 3306
        self.user = db_user or config.adapter.db_user
        self.pswd = db_pass or config.adapter.db_pass
        self.db = db_name or config.adapter.db_name
        self.autocommit = autocommit
        kwargs = dict(
            user=self.user,
            password=self.pswd,
            host=self.host,
            port=int(self.port),
            database=self.db,
            autocommit=self.autocommit,
        )
        self.conn = mariadb.connect(**kwargs)

    def finalize(self):
        self.conn.close()

    def _initialize_tables(self):
        cur = self.conn.cursor()
        cur.execute(self.user_table)
        cur.close()

    @contextmanager
    def adapter_context(self):
        cur = self.conn.cursor()
        try:
            yield cur
        finally:
            cur.close()


class MariaDBQuotaAdapter(PolicyConfigAdapter):
    """A class for adapting to MariaDB for quota policy data"""

    quota_table = (
        "CREATE TABLE IF NOT EXISTS quotas ("  # pragma: no cover
        "id BIGINT AUTO_INCREMENT PRIMARY KEY,"
        "name VARCHAR(32) UNIQUE NOT NULL,"
        "quota BIGINT UNIQUE NOT NULL"
        ")"
    )
    join_table = (
        "CREATE TABLE IF NOT EXISTS quota_user ("  # pragma: no cover
        "quota_id BIGINT NOT NULL,"
        "user_id BIGINT NOT NULL PRIMARY KEY,"
        "CONSTRAINT fk_user"
        " FOREIGN KEY (user_id) REFERENCES users (id)"
        " ON DELETE CASCADE"
        " ON UPDATE RESTRICT,"
        "CONSTRAINT fk_quota"
        " FOREIGN KEY (quota_id) REFERENCES quotas (id)"
        " ON DELETE CASCADE"
        " ON UPDATE CASCADE"  # allow replacement of quota defs
        ")"
    )
    basic_quotas = (
        "INSERT INTO quotas ( name, quota ) VALUES "  # pragma: no cover
        "('10eph', 240),"
        "('50eph', 1200),"
        "('200eph', 4800)"
    )
    quota_query = (
        "SELECT quota FROM quotas WHERE id = ("  # pragma: no cover
        "SELECT quota_id FROM quota_user AS j"
        " LEFT JOIN users AS u ON j.user_id = u.id"
        " WHERE u.name = %(user)s"
        ")"
    )
    quota_map_query = (
        "SELECT u.name AS user, p.quota FROM quotas AS p"  # pragma: no cover
        " LEFT JOIN quota_user AS j ON p.id = j.quota_id"
        " LEFT JOIN users AS u ON j.user_id = u.id"
    )
    quota_map_where = "WHERE u.name IN ({srch})"  # pragma: no cover

    def _initialize_tables(self, *, defquotas=False):
        super()._initialize_tables()
        cur = self.conn.cursor()
        cur.execute(self.quota_table)
        cur.execute(self.join_table)
        if defquotas:
            cur.execute("SELECT COUNT(name) FROM quotas")
            if cur.fetchone()[0] == 0:
                cur.execute(self.basic_quotas)
        cur.close()

    def quota_for_user(self, user):
        """Return the quota for an user account"""
        cur = self.conn.cursor()
        cur.execute(self.quota_query, dict(user=user))
        try:
            res = cur.fetchone()[0]
        except TypeError:  ### generally meaning no result; we could log this
            res = None
        except mariadb.Error as e:  # pragma: no cover
            logger.error(e)  ### CUSTOMIZE: log-level
            res = None
        finally:
            cur.close()
        return res

    def _quota_search(self, users=[]):
        cur = self.conn.cursor()
        if len(users) == 0:
            query = self.quota_map_query
            cur.execute(query)
        else:
            query = self.quota_map_query + " " + self.quota_map_where
            srch = str(users)[1:-1]  ### produces a string suitable for SQL "IN ()"
            cur.execute(query.format(srch=srch))
        return cur.fetchall()

    def quota_dict(self, users=[]):
        """Return a dict which maps users onto their quotas"""
        rows = self._quota_search(users)
        res = {r[0]: r[1] for r in rows}
        return res

    def quota_map(self, func, users=[]):
        """Provide a function to execute over each user and its quota.  Use this to directly wire the database-loading logic to the Redis-population logic."""
        if not callable(func):
            raise ValueError(
                "The first non-self argument must be a callable which accepts the user and quota as arguments, in that order."
            )
        rows = self._quota_search(users)
        res = []
        for row in rows:
            res.append(func(row[0], row[1]))
        return res


class MariaDBSenderDomainAuthAdapter(PolicyConfigAdapter):
    """A class for adapting to MariaDB for sender domain authorization data"""

    domain_table = (
        "CREATE TABLE IF NOT EXISTS domains ("  # pragma: no cover
        "id BIGINT AUTO_INCREMENT PRIMARY KEY,"
        "name VARCHAR(64) UNIQUE NOT NULL"
        ")"
    )
    join_table = (
        "CREATE TABLE IF NOT EXISTS domain_user ("  # pragma: no cover
        "domain_id BIGINT NOT NULL,"
        "user_id BIGINT NOT NULL,"
        "PRIMARY KEY (domain_id, user_id),"  # comp. primary key allows more than one user per domain
        "CONSTRAINT fk_user"
        " FOREIGN KEY (user_id) REFERENCES users (id)"
        " ON DELETE CASCADE"
        " ON UPDATE RESTRICT,"
        "CONSTRAINT fk_domain"
        " FOREIGN KEY (domain_id) REFERENCES domains (id)"
        " ON DELETE CASCADE"
        " ON UPDATE CASCADE"  # allow replacement of domain defs
        ")"
    )
    domain_query = (
        "SELECT d.name FROM domains AS d"  # pragma: no cover
        " WHERE id = ("
        "SELECT domain_id FROM domain_user AS j"
        " LEFT JOIN users AS u ON j.user_id = u.id"
        " WHERE u.name = %(user)s"
        ")"
    )
    domain_map_query = (
        "SELECT u.name AS user, d.name AS domain FROM domains AS d"  # pragma: no cover
        " LEFT JOIN domain_user AS j ON p.id = j.domain_id"
        " LEFT JOIN users AS u ON j.user_id = u.id"
    )
    domain_map_where = "WHERE user IN ({srch})"  # pragma: no cover
    check_domain_query = (
        "SELECT COUNT(d.name) FROM domains AS d"
        " LEFT JOIN domain_user AS j ON d.id = j.domain_id"
        " LEFT JOIN users AS u ON u.id = j.user_id"
        " WHERE d.name = '{domain}' AND u.name = '{user}'"
    )

    def _initialize_tables(self, *args, **kwargs):
        super()._initialize_tables()
        cur = self.conn.cursor()
        cur.execute(self.domain_table)
        cur.execute(self.join_table)
        cur.close()

    def check_domain_for_user(self, user, domain):
        """Returns true if the user is authorized to send for this domain"""
        cur = self.conn.cursor()
        cur.execute(self.check_domain_query.format(user=user, domain=domain))
        result = cur.fetchone()[0]  ### returns 1 if a domain matched, 0 if not
        cur.close()
        return result

    ### These methods would provide a similar functionality profile to the above
    ###     but as yet, none of that code is being used either, so these are left
    ###     as a suggestion for future expansion, as needed
    def _domain_search(self, users=[]):
        """Return a set of rows, either for specified users or all users"""
        raise NotImplementedError

    def domains_for_user(self, user):
        """Return a list of domains the user is authorized to send from"""
        raise NotImplementedError

    def domain_re_for_user(self, user):
        """Returns a regular expression useful for detecting a user's domains"""
        raise NotImplementedError

    def domain_map(self, func, users=[]):
        """Pass in a closure which accepts (user, domain); useful for prepopulating Redis"""
        raise NotImplementedError



"""Signal handlers for CHAPPS"""
import signal, asyncio
import logging
from chapps.config import config

logger = logging.getLogger(__name__)


class SignalHandlerFactory:  # pragma: no cover
    """A class for containing classmethods to create signal handlers"""

    @classmethod
    def signal_handler(cls, loop):
        """Pass in the asyncio event loop and get a closure which will end the program"""

        def signal_handler_closure(sig, frame=None):
            if sig in {signal.SIGTERM, sig.SIGINT}:
                logger.info(f"CHAPPS exiting on {signal.Signals(sig)} ({sig}).")
                raise SystemExit

        return signal_handler_closure


class CHAPPSException(Exception):
    """Parent class for CHAPPS exceptions"""


class CallableExhausted(CHAPPSException):
    """A special exception for use during testing"""


class OutboundPolicyException(CHAPPSException):
    """Exceptions which occur during outbound mail processing"""


class NotAnEmailAddressException(CHAPPSException):
    """The string in question is not an email address"""


class TooManyAtsException(CHAPPSException):
    """An email address had too many at-signs in it"""


class ConfigurationError(CHAPPSException):
    """There was an error in the setting of configuration elements"""


class NullSenderException(OutboundPolicyException):
    """No sender address exists in the current policy request"""


class AuthenticationFailureException(OutboundPolicyException):
    """lack of user_key being treated as authentication failure"""



"""CHAPPS

Caching, Highly-Available Postfix Policy Service

This package contains the following modules:
    util -- utility functions
"""
from ._version import __version__
import chapps.logging

__all__ = [
    "util",
    "config",
    "signals",
    "logging",
    "actions",
    "adapter",
    "policy",
    "spf_policy",
    "outbound",
    "switchboard",
]



"""SQLAlchemy ORM Models to correspond to Pydantic models for API"""
from typing import List
from sqlalchemy import (
    Column,
    Integer,
    String,
    ForeignKey,
    Table,
    select,
    update,
    tuple_,
    delete,
)
from sqlalchemy.orm import (
    declarative_base,
    relationship,
    backref,
    DeclarativeMeta,
)
import logging
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)


class JoinAssoc:
    """
    A class for representing joins via jointable
    There is probably some way to use SQLA for this
    but I am using it wrongly, to avoid loading associations
    just in order to link them to other objects
    via a join table
    """

    def __init__(
        self,
        source_model,
        source_id: str,
        *,
        assoc_name: str,
        assoc_type,
        assoc_model,
        assoc_id: str,
        table,
    ):
        self.source_model = source_model
        self.source_id = source_id
        self.assoc_name = assoc_name
        self.assoc_type = assoc_type
        self.assoc_model = assoc_model
        self.assoc_id = assoc_id
        self.table = table

    def __repr__(self):
        return (
            f"JoinAssoc({self.source_model.__name__}, "
            f"{self.assoc_model.__name__}, "
            f"{self.assoc_name}, {self.assoc_type!r})"
        )

    def select(self):
        return self.table.select()

    def insert(self):
        return self.table.insert()

    def delete(self):
        return self.table.delete()

    def update(self):
        return self.table.update()

    def values(self, item, assoc):
        item_id = item if type(item) == int else item.id
        try:
            i = iter(assoc)
            ins = [{self.source_id: item_id, self.assoc_id: val} for val in i]
        except TypeError:
            ins = [{self.source_id: item_id, self.assoc_id: assoc}]
        return ins

    def where_tuples(self, item_id: int, assoc):
        try:
            i = iter(assoc)
            res = [(item_id, val) for val in i]
        except TypeError:
            res = [(item_id, assoc)]
        return res

    @property
    def source_col(self):
        return getattr(self.table.c, self.source_id)

    @property
    def assoc_col(self):
        return getattr(self.table.c, self.assoc_id)

    def insert_assoc(self, item_id: int, vals):
        return (
            self.insert()
            .prefix_with("IGNORE")
            .values(self.values(item_id, vals))
        )

    def delete_assoc(self, item_id: int, vals):
        return self.delete().where(
            tuple_(self.source_col, self.assoc_col).in_(
                self.where_tuples(item_id, vals)
            )
        )

    def update_assoc(self, item_id: int, assoc_id: int):
        return (
            self.update()
            .where(self.source_col == item_id)
            .values(**{self.assoc_id: assoc_id})
        )

    def select_ids_by_source_id(self, item_id: int):
        # not self.select() b/c we want the SQLA select()
        return select(self.assoc_col).where(self.source_col == item_id)


# declare subclass of the SQLAlchemy DeclarativeMeta class
# in order to attach custom routines to the ORM objects
class DB_Customizations(DeclarativeMeta):
    """Metaclass for adding custom code to ORM models"""

    def select_by_id(cls, id: int):
        return select(cls).where(cls.id == id)

    def windowed_list_by_ids(
        cls,
        *,
        ids: List[int] = [],
        subquery=None,
        skip: int = 0,
        limit: int = 1000,
        q: str = "%",
    ):
        """
        Provide either a list of IDs (integers) or a SELECT which yields them.
        Optionally provide pagination and name filtering.
        """
        if subquery is not None:
            stmt = select(cls).where(cls.id.in_(subquery))
        elif ids:
            stmt = select(cls).where(tuple_(cls.id).in_([(i,) for i in ids]))
        else:
            raise ValueError("Supply one of ids or substatement.")
        stmt = (
            stmt.where(cls.name.like(q))
            .offset(skip)
            .limit(limit)
            .order_by(cls.id)
        )
        return stmt

    def select_names_by_id(cls, ids: List[int]):
        return select(cls.name).where(tuple_(cls.id).in_([(i,) for i in ids]))

    def select_by_pattern(cls, q: str):
        return select(cls).where(cls.name.like(q))

    def select_by_name(cls, q: str):
        return select(cls).where(cls.name == q)

    def windowed_list(cls, q: str = "%", skip: int = 0, limit: int = 1000):
        return (
            cls.select_by_pattern(q).offset(skip).limit(limit).order_by(cls.id)
        )

    def remove_by_id(cls, ids: List[int]):  # (i,) creates a tuple w/ 1 element
        if type(ids) == int:
            ids = [ids]
        return delete(cls).where(tuple_(cls.id).in_([(i,) for i in ids]))

    def update_by_id(cls, item):
        print(f"Got item:{item!r}")
        args = {
            k: getattr(item, k) for k in item.schema()["properties"].keys()
        }
        id = args.pop("id")
        return update(cls).where(cls.id == id).values(**args)


# declare DB model base class
DB_Base = declarative_base(metaclass=DB_Customizations)
"""DB_Base serves as the base of all ORM classes"""

quota_user = Table(
    "quota_user",
    DB_Base.metadata,
    Column(
        "user_id",
        ForeignKey("users.id", ondelete="CASCADE", onupdate="RESTRICT"),
        primary_key=True,
    ),
    Column(
        "quota_id",
        ForeignKey("quotas.id", ondelete="CASCADE", onupdate="CASCADE"),
        nullable=False,
    ),
)
domain_user = Table(
    "domain_user",
    DB_Base.metadata,
    Column(
        "user_id",
        ForeignKey("users.id", ondelete="CASCADE", onupdate="RESTRICT"),
        primary_key=True,
    ),
    Column(
        "domain_id",
        ForeignKey("domains.id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    ),
)


class Quota(DB_Base):
    """ORM Model for quota definitions"""

    __tablename__ = "quotas"

    id = Column(Integer, primary_key=True)
    name = Column(String(32), unique=True)
    quota = Column(Integer, unique=True)

    def __repr__(self):
        return f"Quota[ORM](id={self.id!r}, name={self.name!r}, quota={self.quota!r})"


class Domain(DB_Base):
    """ORM Model for domain definitions"""

    __tablename__ = "domains"

    id = Column(Integer, primary_key=True)
    name = Column(String(64))

    def __repr__(self):
        return f"Domain[ORM](id={self.id!r}, name={self.name!r})"


class User(DB_Base):
    """ORM model for user entities within CHAPPS"""

    __tablename__ = "users"

    id = Column(Integer, primary_key=True)
    name = Column(String(128), unique=True)
    quota = relationship(
        Quota,
        secondary=quota_user,  # using m2m mainly for optimization
        backref="users",  # wants to be efficient and non-eager
        passive_deletes=True,  # prevent loading during deletion
        single_parent=True,  # only one association to quota
        uselist=False,  # one quota per user
        primaryjoin=id == quota_user.c.user_id,
        secondaryjoin=Quota.id == quota_user.c.quota_id,
    )
    domains = relationship(
        Domain,
        secondary=domain_user,  # really m2m
        backref=backref("users", order_by="User.id"),  # reverse associate
        order_by=Domain.id,  # order by id
        passive_deletes=True,  # protect domains
    )

    def __repr__(self):
        return f"User[ORM](id={self.id!r}, name={self.name!r})"



"""API-related objects for CHAPPS"""
# Note that the API is powered by FastAPI and as such, the main API object
# itself is designed to be executed by uvicorn, probably a uvicorn worker
# coordinated by gunicorn (see FastAPI deployment docs)

from chapps.config import config
from chapps._version import __version__
from chapps.rest.routers import users, domains, quotas, live
from fastapi import FastAPI, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
import logging
import chapps.logging
from pathlib import Path

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)

rest_readme = Path(config.chapps.docpath) / "README-API.md"
desc = rest_readme.open("rt").read()

tags_metadata = [
    dict(
        name="users",
        description=(
            "<h3>Create, list, fetch, update, and delete operations "
            "involving users.</h3><p>Update a user "
            "to assign a new quota.  There are some special routes "
            "for adding/removing just some domains without "
            "having to handle the entire set of associations.</p>"
        ),
    ),
    dict(
        name="domains",
        description=(
            "<h3>Create, list, fetch, update and delete operations for "
            "domains.</h3><p>When creating a domain, "
            "it is possible to specify a list of User IDs to "
            "associate.  There are some special routes for managing"
            " user associations (add/remove arbitrary) without having"
            " to handle the entire set of associations.</p>"
        ),
    ),
    dict(
        name="quotas",
        description=(
            "<h3>Create, list, fetch, update and delete operations for "
            "quotas.</h3><p>Since there are very "
            "few quotas compared to the number of users, there is "
            "no support planned for managing users from their "
            "associated quotas.  To change a quota for a user, "
            "update the user.</p>"
        ),
    ),
    dict(
        name="live",
        description=(
            "<h3>Status reporting and remote command interface.</h3><p>Routes "
            "provided for obtaining real-time remaining quota, resetting "
            "quota, refreshing policy settings, adding or removing SDAs, "
            "and causing CHAPPS to rewrite its config file, in order to "
            "allow for updates post-upgrade without losing site customizations.</p>"
        ),
    ),
]

api = FastAPI(
    title="CHAPPS REST API",
    description=desc,
    version=__version__,
    contact=dict(
        name="Caleb S. Cullen",
        url="https://github.com/easydns/chapps",
        email="ccullen@easydns.com",
    ),
    license_info=dict(name="MIT License", url="https://mit-license.org/"),
    openapi_tags=tags_metadata,
)
verstr = config.chapps.version
api.include_router(users.api)
api.include_router(domains.api)
api.include_router(quotas.api)
api.include_router(live.api)


@api.exception_handler(RequestValidationError)
async def validation_exception_handler(
    request: Request, exc: RequestValidationError
):
    exc_str = f"{exc}".replace("\n", " ").replace("   ", " ")
    body = await request.json()
    logging.error(f"{body!r}: {exc_str}")
    content = {"status_code": 10422, "message": exc_str, "data": None}
    return JSONResponse(
        content=content, status_code=status.HTTP_422_UNPROCESSABLE_ENTITY
    )



from chapps.config import config
from chapps.rest import dbmodels
from typing import Optional, List, Dict
from pydantic import BaseModel
from pydantic.main import ModelMetaclass
from enum import Enum
import time

verstr = config.chapps.version


class AssocOperation(str, Enum):
    """
    'add', 'subtract', or 'set'; use 'set' only with unitary associations
    """

    add = "add"
    subtract = "subtract"
    replace = "replace"


class SDAStatus(str, Enum):
    AUTH = "AUTHORIZED"
    PROH = "PROHIBITED"
    NONE = "NOT CACHED"


# a metaclass for passing calls through to the orm_model
class CHAPPSMetaModel(ModelMetaclass):
    def __getattr__(cls, var):
        orm_method = getattr(cls.Meta.orm_model, var, None)
        return orm_method or super().__getattr__(var)


# there could be a metaclass which would look for dbmodels classes matching
# the names of subclasses of CHAPPSModel and automatically hook up their
# Meta.orm_model data ... if the number of tables starts to grow
class CHAPPSModel(BaseModel, metaclass=CHAPPSMetaModel):
    """Base API data model"""

    id: int
    name: str

    class Meta:
        orm_model = dbmodels.DB_Base

    @classmethod
    def id_name(cls):
        return "_".join(str(cls.Meta.orm_model.id).lower().split("."))

    @classmethod
    def wrap(cls, orm_instance):
        """create a pydantic model from an ORM model"""
        if not orm_instance:  # could be None or []
            return orm_instance
        try:
            orm_iter = iter(orm_instance)
            return [cls.from_orm(oi) for oi in orm_iter]
        except TypeError:
            return cls.from_orm(orm_instance)

    @classmethod
    def join_assoc(cls, **kwargs):
        # if it became necessary to track some other arbitrary id-column name
        # we could accomplish that with a metaclass, and then just set it
        # in each subclass
        return dbmodels.JoinAssoc(cls, cls.__name__.lower() + "_id", **kwargs)


class User(CHAPPSModel):
    """API model to represent users"""

    class Config:
        orm_mode = True
        schema_extra = dict(
            example=dict(id=0, name=("[user.identifier@]domain.name"))
        )

    class Meta:
        orm_model = dbmodels.User


class Quota(CHAPPSModel):
    """API model to represent quotas"""

    quota: int

    class Config:
        orm_mode = True
        schema_extra = dict(
            example=dict(id=0, name="fiftyPerHour", quota=1200)
        )

    class Meta:
        orm_model = dbmodels.Quota


class Domain(CHAPPSModel):
    """A model to represent domains"""

    class Config:
        orm_mode = True
        schema_extra = dict(example=dict(id=0, name="[sub.]domain.tld"))

    class Meta:
        orm_model = dbmodels.Domain


class CHAPPSResponse(BaseModel):
    version: str
    timestamp: float
    response: object

    class Config:
        schema_extra = dict(version="CHAPPS v0.4", timestamp=time.time())

    @classmethod
    def send(model, response=None, **kwargs):
        """Utility function for encapsulating responses in a standard body"""
        mkwargs = dict(version=verstr, timestamp=time.time())
        if response:
            mkwargs["response"] = response
        return model(**mkwargs, **kwargs)


class UserResp(CHAPPSResponse):
    response: User
    domains: Optional[List[Domain]] = None
    quota: Optional[Quota] = None


class UsersResp(CHAPPSResponse):
    response: List[User]


class DomainResp(CHAPPSResponse):
    response: Domain
    users: Optional[List[User]] = None


class DomainsResp(CHAPPSResponse):
    response: List[Domain]


class QuotaResp(CHAPPSResponse):
    response: Quota


class QuotasResp(CHAPPSResponse):
    response: List[Quota]


class IntResp(CHAPPSResponse):
    response: int


class TextResp(CHAPPSResponse):
    response: str


class LiveQuotaResp(CHAPPSResponse):
    response: int
    remarks: List[str] = []


class DomainUserMapResp(CHAPPSResponse):
    """
    A dict-of-dicts:
      top level key is domain name
      second key is user
      value is SDAStatus
    """

    response: Dict[str, Dict[str, SDAStatus]]


class DeleteResp(TextResp):
    response: str = "deleted"


### the following classes are somewhat speculative for now
class FloatResp(CHAPPSResponse):
    response: float



#!/usr/bin/env python3
"""CHAPPS database setup script"""
### This script reads the CHAPPS config and uses the database information contained therein
###   to initialize the database.  It should be idempotent, and safe to run against an existing
###   config database.

from chapps.adapter import MariaDBQuotaAdapter, MariaDBSenderDomainAuthAdapter


def setup_chapps_database():
    for adapter in [MariaDBQuotaAdapter(), MariaDBSenderDomainAuthAdapter()]:
        adapter._initialize_tables()


if __name__ == "__main__":
    setup_chapps_database()



"""Classes related to outbound traffic"""
from chapps.util import PostfixPolicyRequest
from chapps.config import config
from chapps.signals import ConfigurationError, AuthenticationFailureException
import logging

logger = logging.getLogger(__name__)


class OutboundPPR(
    PostfixPolicyRequest
):  # empty line eliminated for paste-ability
    memoized_routines = dict()
    # Initialize with optional config, in order to allow site-specific user-search path
    def __init__(self, payload, *, cfg=None):
        super().__init__(payload)
        self._config = cfg or config
        self._config = self._config.chapps

    @classmethod
    def clear_memoized_routines(cls):
        cls.memoized_routines.clear()

    def __str__(self):
        try:
            return (f"i={self.instance} "
                    f"user={self.user} "
                    f"sender={self.sender or 'None'} "
                    f"client_address={self.client_address} "
                    f"recipient={self.recipient}")
        except Exception:
            return (f"i={self.instance} "
                    f"sasl_username={self.sasl_username or 'None'} "
                    f"ccert_subject={self.ccert_subject or 'None'} "
                    f"sender={self.sender or 'None'} "
                    f"client_address={self.client_address} "
                    f"recipient={self.recipient}")

    ### create a property handler for "user"
    @property
    def user(self):
        if not "_user" in self.__dict__:
            try:
                self._user = self._get_user()
            except ValueError as e:
                if self._config.require_user_key:
                    raise AuthenticationFailureException()
                else:
                    raise e
        return self._user

    # this routine creates and memoizes a routine if need be
    # it also uses that routine to return the value to be used to track a user
    def _get_user(self):
        """Obtain the user value, and memoize the procedure"""
        # see if we already have a procedure from some previous iteration
        get_user = self.__class__.memoized_routines.get("get_user", None)
        cfg = self._config
        # if there is no procedure, we build one
        if not get_user:
            if cfg.require_user_key:
                if not cfg.user_key:
                    raise ConfigurationError(
                        (
                            "If require_user_key is True, "
                            "then user_key must be set."
                        )
                    )
                qk_list = [cfg.user_key]
            else:
                qk_list = [
                    "sasl_username",
                    "ccert_subject",
                    "sender",
                    "client_address",
                ]
                qk = cfg.user_key
                if qk and qk != qk_list[0]:
                    qk_list = [qk, *qk_list]

            def get_user(ppr):
                for k in qk_list:
                    user = getattr(ppr, k, None)
                    if user and user != "None":
                        logger.debug(
                            f"Selecting user-identifier {user} from key {k}"
                        )
                        return user
                raise ValueError(
                    (
                        "None of the following keys had values in "
                        f"the provided PPR: {qk_list}"
                    )
                )

            # this memoizes the procedure for finding the user
            self.__class__.memoized_routines["get_user"] = get_user
        # execute the procedure and return the result
        return get_user(self)



"""SPF Enforcement Policy Object"""
import spf
from chapps.policy import EmailPolicy
from chapps.actions import PostfixSPFActions


class SPFEnforcementPolicy(EmailPolicy):
    """Represents SPF policy"""

    ### we may never use Redis for SPF directly
    redis_key_prefix = "spf"

    def __init__(self, cfg=None):
        """At present, there are no extra options for SPF Enforcement policy; perhaps there should be for
           how to treat none_neutral and softfail"""
        super().__init__(cfg)
        self.actions = PostfixSPFActions()

    def approve_policy_request(self, ppr):
        """Perform SPF enforcement decision-making on the PPR"""
        ### First, check the HELO name
        helo_sender = "postmaster@" + ppr.helo_name
        query = spf.query(ppr.client_address, helo_sender, ppr.helo_name)
        result, _, message = query.check()
        if result in ["fail"]:  ### TODO: allow configuration of HELO results to honor
            action = self.actions.action_for(result)
        else:
            ### the HELO name did not produce a definitive result, so check MAILFROM
            query = spf.query(ppr.client_address, ppr.sender, ppr.helo_name)
            result, _, message = query.check()
            action = self.actions.action_for(result)
        return action(message, ppr=ppr, prepend=query.get_header(result))



from typing import List
from starlette import status
from fastapi import APIRouter  # , Body, Path, HTTPException
from chapps.rest.models import (
    User,
    Domain,
    DomainResp,
    DomainsResp,
    UsersResp,
    DeleteResp,
    TextResp,
    AssocOperation,
)
from .common import (
    get_item_by_id,
    list_items,
    create_item,
    delete_item,
    update_item,
    adjust_associations,
    list_associated,
)
import logging
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)

api = APIRouter(
    prefix="/domains",
    tags=["domains"],
    responses={
        status.HTTP_404_NOT_FOUND: {"description": "Domain not found."}
    },
)

domain_join_assoc = [
    Domain.join_assoc(
        assoc_name="users",
        assoc_type=List[int],
        assoc_model=User,
        assoc_id="user_id",
        table=Domain.Meta.orm_model.metadata.tables["domain_user"],
    )
]

api.get("/", response_model=DomainsResp)(
    list_items(Domain, response_model=DomainsResp)
)

api.get("/{item_id}", response_model=DomainResp)(
    get_item_by_id(Domain, response_model=DomainResp, assoc=[(User, "users")])
)

api.get("/{item_id}/allowed/", response_model=UsersResp)(
    list_associated(
        Domain, assoc=domain_join_assoc[0], response_model=UsersResp
    )
)

api.post(
    "/",
    status_code=201,
    response_model=DomainResp,
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "description": "Unable to create domain"
        },
        status.HTTP_409_CONFLICT: {"description": "Unique key error."},
    },
)(
    create_item(
        Domain,
        response_model=DomainResp,
        params=dict(name=str),
        assoc=domain_join_assoc,
    )
)

api.put("/", response_model=DomainResp)(
    update_item(Domain, response_model=DomainResp, assoc=domain_join_assoc)
)

api.put("/{item_id}/allow/", response_model=TextResp)(
    adjust_associations(
        Domain, assoc=domain_join_assoc, assoc_op=AssocOperation.add
    )
)

api.put("/{item_id}/deny/", response_model=TextResp)(
    adjust_associations(
        Domain, assoc=domain_join_assoc, assoc_op=AssocOperation.subtract
    )
)

api.delete(
    "/",
    response_model=DeleteResp,
    responses={
        status.HTTP_202_ACCEPTED: {"description": "Items will be deleted."},
        status.HTTP_204_NO_CONTENT: {"description": "No item to delete."},
        status.HTTP_409_CONFLICT: {
            "description": "Database integrity conflict."
        },
    },
)(
    delete_item(Domain)
)  # params=dict(ids=List[int]) by default



"""chapps.policy

Policy routines for CHAPPS
"""
import time
from contextlib import contextmanager
from collections import deque
import functools
import redis
import logging
from expiring_dict import ExpiringDict
from chapps.config import config
from chapps.adapter import MariaDBQuotaAdapter, MariaDBSenderDomainAuthAdapter
from chapps.signals import (
    TooManyAtsException,
    NullSenderException,
    NotAnEmailAddressException,
)
from chapps.rest.models import Quota, SDAStatus

logger = logging.getLogger(__name__)
seconds_per_day = 3600 * 24
SENTINEL_TIMEOUT = 0.1
TIME_FORMAT = "%d %b %Y %H:%M:%S %z"


class EmailPolicy:
    redis_key_prefix = "chapps"

    @staticmethod
    def rediskey(prefix, *args):
        """Format a string to serve as a Redis key for arbitrary data"""
        return f"{ prefix }:{ ':'.join( args ) }"

    @classmethod
    def _fmtkey(cls, *args):
        return cls.rediskey(cls.redis_key_prefix, *args)

    def __init__(self, cfg=None):
        self.config = cfg if cfg else config
        self.params = self.config.get_block(self.__class__.__name__)
        self.sentinel = None
        self.redis = self._redis()  # pass True to get read-only
        self.instance_cache = ExpiringDict(3)  # entries expire after 3 seconds

    def _redis(self, read_only=False):
        """Get a Redis handle, possibly from Sentinel"""
        try:
            if self.config.redis.sentinel_servers and not self.sentinel:
                self.sentinel = redis.Sentinel(
                    [
                        s.split(":")
                        for s in self.config.redis.sentinel_servers.split(" ")
                    ],
                    socket_timeout=SENTINEL_TIMEOUT,
                )
            if self.sentinel:
                if read_only:
                    rh = self.sentinel.slave_for(
                        self.config.redis.sentinel_dataset,
                        socket_timeout=SENTINEL_TIMEOUT,
                    )
                else:
                    rh = self.sentinel.master_for(
                        self.config.redis.sentinel_dataset,
                        socket_timeout=SENTINEL_TIMEOUT,
                    )
                return rh
        except AttributeError:
            pass
        return redis.Redis(
            host=self.config.redis.server, port=self.config.redis.port
        )

    def approve_policy_request(self, ppr):
        raise NotImplementedError(
            "Subclasses of EmailPolicy must implement this function."
        )


class GreylistingPolicy(EmailPolicy):
    """Represents Greylisting policy"""

    redis_key_prefix = "grl"

    def __init__(
        self,
        cfg=None,
        *,
        minimum_deferral=60,
        cache_ttl=seconds_per_day,
        auto_allow_after=10,
    ):
        """all EmailPolicy objects expect an optional positional param for a CHAPPSConfig
           named arguments for GreylistingPolicy objects are:
             - minimum_deferral: the minimum amount of time allowed for greylisting to approve the second attempt
             - cache_ttl: how long to store data about a tracked client (source IP)
                          this is used both for auto-allow and for basic greylist impression retention
             - auto_allow_after: a count of successful greylist re-sends post-deferral, after which
                          approval will be granted automatically
        """
        super().__init__(cfg)
        self.min_defer = minimum_deferral
        self.cache_ttl = cache_ttl
        self.allow_after = auto_allow_after
        if self.cache_ttl <= self.min_defer:
            logger.warning(
                f"Cache TTL (={datetime.timedelta(seconds=self.cache_ttl)}) is not allowed to be smaller than or equal to the minimum deferral window (={datetime.timedelta(seconds=self.min_defer)}).  Defaulting to 24 hr."
            )
            self.cache_ttl = seconds_per_day
        if self.min_defer > 60 * 15:
            logger.warning(
                f"It may be unreasonable to expect the sending server to defer for more than 15 minutes. (={self.min_defer/60.0:.2f}m)"
            )
        if self.allow_after == 0:
            logger.warning(f"Sender auto-approval is turned off.")
        elif self.allow_after < 2:
            logger.warning(
                f"Sender auto-approval is set to a fairly low threshold. (={self.allow_after})"
            )

    def tuple_key(self, ppr):
        """The names of the values in the PPR are as follows (in order):
             - client_address
             - sender
             - recipient
        """
        return self._fmtkey(ppr.client_address, ppr.sender, ppr.recipient)

    def client_key(self, ppr):
        return self._fmtkey(ppr.client_address)

    def approve_policy_request(self, ppr):
        """Expects a PostfixPolicyRequest; returns True if the email should be accepted"""
        ### TODO: memoization can be factored into the superclass if pre-evaluation hook and evaluation
        ###       can be generalized
        instance = ppr.instance
        cached_response = self.instance_cache.get(instance, None)
        if cached_response is not None:
            return cached_response
        response = self._evaluate_policy_request(ppr)
        self.instance_cache[instance] = response
        return response

    def _evaluate_policy_request(self, ppr):
        try:
            logger.debug(f"Getting control data for {self.tuple_key( ppr )}")
            tuple_seen, client_tally = self._get_control_data(ppr)
            logger.debug(f"Got values ({tuple_seen}, {client_tally})")
        except Exception:  # pragma: no cover
            logger.exception("UNEXPECTED")
            logger.debug(
                f"Returning denial for {ppr.instance} (unexpected exception)."
            )
        if client_tally is not None and client_tally >= self.allow_after:
            self._update_client_tally(ppr)
            return True
        if tuple_seen:
            # The tuple is recognized; need to check if it was long enough ago
            now = time.time()
            if now - tuple_seen >= self.min_defer:
                # the email will be approved; some housekeeping is necessary
                self._update_client_tally(ppr)
                return True
        # if we get here, the tuple either isn't stored or was stored too recently
        # either way, we update it
        self._update_tuple(ppr)
        return False

    def _get_control_data(self, ppr):
        """Extract data from Redis in order to answer the policy request"""
        now = time.time()
        client_address, sender, recipient = (
            ppr.client_address,
            ppr.sender,
            ppr.recipient,
        )
        tuple_key = self.tuple_key(ppr)
        client_key = self.client_key(ppr)
        pipe = self.redis.pipeline()
        pipe.zremrangebyscore(client_key, 0, now - float(self.cache_ttl))
        pipe.get(tuple_key)
        if self.allow_after > 0:
            pipe.zrange(client_key, 0, -1)

        result = pipe.execute()
        tuple_bits = result[1]
        if len(result) == 3:
            client_tally_bits = result[2]

        tuple_seen = (
            float(tuple_bits) if tuple_bits else None
        )  # UNIX epoch time
        client_tally = None
        if self.allow_after > 0 and client_tally_bits:
            client_tally = len(client_tally_bits)
        return (tuple_seen, client_tally)

    def _update_client_tally(self, ppr):
        """Update Redis when an email is allowed, to increase the reliability score of the client"""
        if self.allow_after == 0:  # if we're not keeping a tally, return
            return
        now = time.time()
        client_key = self.client_key(ppr)
        with self.redis.pipeline() as pipe:
            pipe.zadd(client_key, {ppr.instance: now})
            pipe.zremrangebyrank(
                client_key, 0, -(self.allow_after + 2)
            )  # keep one extra
            pipe.expire(client_key, self.cache_ttl)
            pipe.execute()

    def _update_tuple(self, ppr):
        self.redis.setex(self.tuple_key(ppr), self.cache_ttl, time.time())


class OutboundQuotaPolicy(EmailPolicy):
    """Represents an outbound quota policy, based on sending rate"""

    redis_key_prefix = "oqp"

    def __init__(self, cfg=None, *, enforcement_interval=None, min_delta=0):
        """first, optional positional argument: a CHAPPSConfig object to use
           named arguments: enforcement_interval will default to seconds per day if not provided
                            min_delta defaults to 5 seconds, to prevent spamming; set to 0 to disable
        """
        super().__init__(cfg)  # sets attrs 'config', 'params', and 'redis'
        self.interval = (
            enforcement_interval if enforcement_interval else seconds_per_day
        )
        if hasattr(self.params, "min_delta"):
            self.min_delta = self.params.min_delta
        elif min_delta:
            self.min_delta = min_delta
        else:
            self.min_delta = 0
        self.min_delta = float(self.min_delta)
        self.counting_recipients = (
            self.params.counting_recipients
            if hasattr(self.params, "counting_recipients")
            else False
        )

    @contextmanager
    def _control_data_storage_context(self):
        """return a closure which takes the tuple (email, quota, margin) and stores it in Redis"""
        pipe = self.redis.pipeline()
        fmtkey = self._fmtkey

        def _dsc(user, quota, margin):
            pipe.set(fmtkey(user, "limit"), quota, ex=seconds_per_day)
            pipe.set(fmtkey(user, "margin"), margin, ex=seconds_per_day)

        try:
            yield _dsc
        finally:
            pipe.execute()
            pipe.reset()

    @contextmanager
    def _adapter_handle(self):
        """A context manager for obtaining a database handle for acquiring policy data"""
        adapter = MariaDBQuotaAdapter(
            db_host=self.config.adapter.db_host,
            db_port=self.config.adapter.db_port,
            db_name=self.config.adapter.db_name,
            db_user=self.config.adapter.db_user,
            db_pass=self.config.adapter.db_pass,
        )
        try:
            yield adapter
        finally:
            adapter.conn.close()

    def _get_control_data(self, ppr):
        """
        This is the routine which keeps track of emails in Redis.  It
        combines all of its requests into a single pipelined (atomic)
        transaction.  When counting recipients, the record is a string
        consisting of the timestamp and the recipient serial number
        separated by a colon, in order to ensure that each recipient
        is listed as an attempt in the log.  The score is always the
        floating-point return value of time.time()
        """
        # cache the current(ish) time
        time_now = time.time()
        time_now_s = str(time_now)

        user = ppr.user

        # create a dict for Redis.zadd()
        if self.counting_recipients:
            tries_dict = {
                (time_now_s + f":{i:05d}"): time_now
                for i, r in enumerate(ppr.recipients)
            }
        else:
            tries_dict = {time_now_s: time_now}
        # set up the Redis keys
        tries_key = self._fmtkey(user, "attempts")
        limit_key = self._fmtkey(user, "limit")
        margin_key = self._fmtkey(user, "margin")
        # Create a Redis pipeline to atomize a set of instructions
        pipe = self.redis.pipeline()
        # Clear the list down to just the last interval seconds, generally a day
        pipe.zremrangebyscore(tries_key, 0, time_now - float(self.interval))
        # Add this/these try(es)
        pipe.zadd(tries_key, tries_dict)
        # Get control data: the limit, the margin, the attempts list
        pipe.get(limit_key)
        pipe.get(margin_key)
        pipe.zrange(tries_key, 0, -1)
        # Set expires on all this stuff so that if they don't send
        # email for a day, their data won't still be sitting around
        # takin' up space; these return nil values so we have to
        # ignore them when we get the results
        pipe.expire(tries_key, self.interval)
        pipe.expire(limit_key, self.interval)
        pipe.expire(margin_key, self.interval)
        # Do the thing!
        results = pipe.execute()
        # The non-retrieval operations still have return values, so we ignore them
        removed, _, limit, margin, attempts, _, _, _ = results
        # Always polite to reset your pipe
        pipe.reset()
        # Attempt typecasting on the margin number, which is allowed
        # to be either int or float
        m = self._cast_margin(margin)
        # If no limit is defined, that means there is no quota profile
        # for the user, and we just return None here; we must test against
        # None because it might be 0
        return (int(limit) if limit is not None else None, m, attempts)

    def _cast_margin(self, margin_bytes):
        try:
            m = int(margin_bytes)
        except:
            try:
                m = float(margin_bytes)
            except:
                m = 0
        return m

    def current_quota(self, user: str, quota: Quota):
        attempts_key = self._fmtkey(user, "attempts")
        limit_key = self._fmtkey(user, "limit")
        pipe = self.redis.pipeline()
        pipe.zremrangebyscore(
            attempts_key, 0, time.time() - float(self.interval)
        )
        pipe.get(limit_key)
        pipe.zrange(attempts_key, 0, -1)
        results = pipe.execute()
        _, limit_bytes, attempts_bytes = results
        pipe.reset()
        limit = (
            int(limit_bytes)
            if limit_bytes is not None
            else quota.quota
            if quota is not None
            else None
        )
        response = limit - len(attempts_bytes) if limit else 0
        remarks = []
        if attempts_bytes:
            last = attempts_bytes[-1]
            try:
                last = float(last)
            except ValueError:
                last = float(last.split(b":")[0])
            last_try = time.strftime(TIME_FORMAT, time.gmtime(last))
            remarks.append(f"Last send attempt was at {last_try}")
        if not limit_bytes:
            remarks.append(f"There is no cached quota limit for {user}.")
        if not quota:
            remarks.append(f"There is no quota configured for user {user}.")
        if not limit:
            remarks.append(
                f"No limit could be found; returning zero xmits remaining."
            )
        return (response, remarks)

    def reset_quota(self, user: str):
        attempts_key = self._fmtkey(user, "attempts")
        pipe = self.redis.pipeline()
        pipe.zrange(attempts_key, 0, -1)
        pipe.delete(attempts_key)
        results = pipe.execute()
        pipe.reset()
        attempts = results[0]
        if attempts:
            msg = f"Attempts (quota) reset for {user}:"
            n_att = len(attempts)
        else:
            n_att = 0
            msg = f"No attempts to reset for {user}:"
        msg += f" {n_att} xmits dropped"
        return (n_att, [msg])

    def refresh_policy_cache(self, user: str, quota: Quota):
        """API adapter method for refreshing policy config cache"""
        self.acquire_policy_for(user, quota.quota)
        return self.current_quota(user, quota)

    def _store_control_data(self, user, quota, margin=0):
        """Using a context manager, build up a set of instructions to store control data"""
        with self._control_data_storage_context() as dsc:
            if type(margin) == float:
                if margin > 1.0:
                    if margin < 100.0:
                        margin = margin / 100.0
                    else:
                        raise TypeError(
                            "margin must be a positive integer or a positive float less than 1 (a percentage)"
                        )  # pragma: no cover
                margin = int(margin * float(quota))
            dsc(user, quota, margin)

    def _detect_control_data(self, user):
        """See if there is control data in Redis for a particular sender"""
        key = self._fmtkey(user, "limit")
        res = None
        try:
            res = self.redis.get(key)
        except redis.exceptions.ResponseError:  # pragma: no cover
            # sometimes a key which should not exist still shows up,
            # and accessing them causes this error.  we choose to
            # pretend nothing happened and just delete the key
            self.redis.delete(key)
        return res

    def acquire_policy_for(self, user, quota=None):
        """Go get the policy for a sender from the relational database or
           other policy source via adapter.  If the margin needs to be
           altered on a per-sender basis, this is the place to adjust
           that.
        """
        if not quota:
            with self._adapter_handle() as adapter:
                quota = adapter.quota_for_user(user)
        if quota:
            self._store_control_data(user, quota, self.params.margin)

    def approve_policy_request(self, ppr):
        """Expects a PostfixPolicyRequest object; returns True if mail sending
           should be allowed This routine implements memoization in
           order to overcome the Postfix double-checking weirdness.
           Since it seems like the object only lives for the duration
           of the query, it doesn't get us very far.  The Postfix docs
           say it will hold open and re-use the connection, but it
           does not seem to.
        """
        user = ppr.user
        instance = ppr.instance
        cached_response = self.instance_cache.get(instance, None)
        if cached_response is not None:
            return cached_response
        if not self._detect_control_data(
            user
        ):  # attempt to retrieve control data if need be
            self.acquire_policy_for(user)
        response = self._evaluate_policy_request(ppr)
        self.instance_cache[instance] = response
        return response

    def _get_delta(self, ppr, attempts):
        """
        This routine gets the number of seconds between this attempt and the
        previous.  It takes into account whether we are counting each
        recipient, and parses the attempt record accordingly -- when counting
        recipients, the record is a string consisting of the timestamp and the
        recipient serial number separated by a colon, in order to ensure that
        each recipient is listed as an attempt in the log.
        """
        if len(attempts) < 2:
            return float("inf")
        delta_index = [-1, -2]
        if not self.counting_recipients:
            if len(attempts) < 2:
                return float("inf")
        elif len(attempts) > len(ppr.recipients):
            # skip back all but one recipient
            recipients_offset = 0 - len(ppr.recipients)
            delta_index = [d + recipients_offset for d in delta_index]
        else:
            logger.debug()
            return float("inf")  # automatically wins
        logger.debug(
            f"Looking at time-delta for {ppr}: indices {delta_index!r}"
        )
        try:
            timestamps = [
                float(t.decode("utf-8").split(":")[0])
                if ":" in t.decode("utf-8")
                else float(t.decode("utf-8"))
                for t in [
                    attempts[i] for i in delta_index if i < len(attempts)
                ]
            ]
        except IndexError:
            msg = (
                f"Recipients={-recipients_offset}"
                f" delta_indices={delta_index!r}"
                f" Attempts: (#{len(attempts)})"
            )
            if len(attempts) < 10:
                msg += f" {attempts!r}"
            logger.exception(msg)
            return float("inf")
        if len(timestamps) == 2:
            logger.debug(
                f"attempts: {[attempts[i] for i in delta_index]}; timestamps: {timestamps!r}"
            )
            return timestamps[0] - timestamps[1]
        return float("inf")  # return a large value

    def _evaluate_policy_request(self, ppr):
        """
        This actually checks to see if it's okay to send the email.
        """
        instance, user = ppr.instance, ppr.user
        try:  # this may raise TypeError if the user is unknown
            limit, margin, attempts = self._get_control_data(ppr)
        except Exception:  # pragma: no cover
            logger.exception("UNEXPECTED")
            logger.debug(
                f"Returning denial indicator for {instance} (unexpected exception)."
            )
            return False
        if not limit:  # user does not have a quota profile
            return False
        if len(attempts) < 2:  # this is the first attempt in the Redis history
            logger.debug(f"Returning OK for {instance} (first attempt).")
            return True
        if self.min_delta != 0:  # set up for checking on throttle
            logger.debug(
                f"Checking throttle: {ppr.user}:{instance} limit: {limit} margin: {margin} tries: {len(attempts)}"
            )
            this_delta = self._get_delta(ppr, attempts)
            if this_delta < float(self.min_delta):  # trying too fast
                logger.debug(
                    f"Rejecting {instance} of {user} for trying too fast. ({this_delta}s since last attempt)"
                )
                return False
        if (
            len(attempts) > limit
        ):  # not too fast, check how many send attempts on record
            ### TODO: alert when the attempts list is really long -- perhaps via Redis pub/sub
            if (
                len(attempts) - margin > limit
                or len(attempts) - len(ppr.recipients) >= limit
            ):
                logger.debug(
                    f"Rejecting {instance} of {user} for having too many attempts in the last interval: recip: {len(ppr.recipients)} limit: {limit}; tries: {len(attempts)}"
                )
                return False
            else:
                logger.debug(
                    f"Returning OK for {instance} recip: {len(ppr.recipients)} limit: {limit}; tries: {len(attempts)} (within margin)."
                )
        logger.debug(f"Returning OK for {instance} (under quota).")
        return True


class SenderDomainAuthPolicy(EmailPolicy):
    """A class for encapsulation of explicit policy regarding what authenticated users are allowed to send from what domains.  Right now we match on the entire string after the @ """

    # every subclass of EmailPolicy must set a key prefix
    redis_key_prefix = "sda"
    # initialization is when we plug in the config
    def __init__(self, cfg=None):
        """first, optional positional argument: a CHAPPSConfig object to use"""
        super().__init__(cfg)  # sets attrs 'config' and 'redis'

    # every subclass has one of these, with a unique name, and fine
    # but maybe there should also be a generic single entry point
    def sender_domain_key(self, ppr):
        """Create a Redis key for each valid user->domain mapping, for speed"""
        return self._sender_domain_key(ppr.user, self._get_sender_domain(ppr))

    # factored out for use in API
    def _sender_domain_key(self, user, domain):
        return self._fmtkey(user, domain)

    # determine the domain of the sender address, if any
    @functools.lru_cache(maxsize=2)
    def _get_sender_domain(self, ppr):
        if ppr.sender:
            parts = ppr.sender.split("@")
            if len(parts) > 2:
                logger.info(
                    "Found sender email with more than one at-sign: "
                    f"sender={ppr.sender} instance={ppr.instance} "
                    f"parts={parts!r}"
                )
                raise TooManyAtsException(f"{ppr.sender}=>{parts!r}")
            elif len(parts) == 1:
                logger.info(
                    "Found sender string without at-sign: "
                    f"sender={ppr.sender} instance={ppr.instance} "
                    f"parts={parts!r}"
                )
                raise NotAnEmailAddressException
            return parts[-1]
        raise NullSenderException

    # We will need to be able to access policy data in the RDBMS
    def _detect_control_data(self, user, domain):
        """Look for SDA control data for a user"""
        key = self._sender_domain_key(user, domain)
        res = None
        try:
            res = self.redis.get(key)
        except redis.exceptions.ResponseError:  # pragma: no cover
            self.redis.delete(key)
        return res

    def _get_control_data(self, ppr):
        return self._detect_control_data(
            ppr.user, self._get_sender_domain(ppr)
        )

    # We will need to be able to store data in Redis
    def _store_control_data(self, ppr, allowed):
        with self._control_data_storage_context() as dsc:
            dsc(ppr.user, self._get_sender_domain(ppr), allowed)

    # We will need a Redis storage context manager in order to mimic the structure of OQP
    @contextmanager
    def _control_data_storage_context(self, expire_time=seconds_per_day):
        pipe = self.redis.pipeline()
        fmtkey = self._fmtkey

        def _dsc(user, domain, allowed):
            pipe.set(fmtkey(user, domain), allowed, ex=seconds_per_day)

        try:
            yield _dsc
        finally:
            pipe.execute()
            pipe.reset()

    # We will need a database adapter context manager
    @contextmanager
    def _adapter_handle(
        self
    ):  # TODO: identical to OQP -- should be factored up into a superclass
        adapter = MariaDBSenderDomainAuthAdapter(
            db_host=self.config.adapter.db_host,
            db_port=self.config.adapter.db_port,
            db_name=self.config.adapter.db_name,
            db_user=self.config.adapter.db_user,
            db_pass=self.config.adapter.db_pass,
        )
        try:
            yield adapter
        finally:
            adapter.conn.close()

    # How to obtain control data
    def acquire_policy_for(self, ppr):
        with self._adapter_handle() as adapter:
            allowed = adapter.check_domain_for_user(
                ppr.user, self._get_sender_domain(ppr)
            )
        if allowed is not None:
            self._store_control_data(ppr, 1 if allowed else 0)
        return allowed

    # This is the main purpose of the class, to answer this question
    def approve_policy_request(self, ppr):
        """
        Given a PPR, say whether this user is allowed to send as the
        apparent sender domain
        """
        result = self.instance_cache.get(
            ppr.instance, None
        )  # instances sometimes repeat
        if not result:
            result = self._get_control_data(ppr)
            if result is None:
                result = self.acquire_policy_for(ppr)
            self.instance_cache[ppr.instance] = result
        return bool(int(result))

    def _decode_policy_cache(self, result):
        if result is not None:
            result = int(result)
            if result:
                return SDAStatus.AUTH
            else:
                return SDAStatus.PROH
        else:
            return SDAStatus.NONE

    # For the API -- inspect state
    def check_policy_cache(self, user, domain):
        return self._decode_policy_cache(
            self._detect_control_data(user, domain)
        )

    def clear_policy_cache(self, user, domain):
        prev = self.check_policy_cache(user, domain)
        if prev != SDAStatus.NONE:
            self.redis.delete(self._sender_domain_key(user, domain))
        return prev

    def bulk_clear_policy_cache(self, users, domains):
        # there seems to be no max pipeline size
        # but if things get sketchy, we can chunk this
        with self.redis.pipeline() as pipe:
            for d in domains:
                for u in users:
                    pipe.delete(self._sender_domain_key(u, d))
            pipe.execute()
            pipe.reset()

    def bulk_check_policy_cache(self, users, domains):
        """Build a map based on domain, full of maps from username to status"""
        with self.redis.pipeline() as pipe:
            for d in domains:
                for u in users:
                    logger.debug(f"bcpc seeking SDA for {u} from {d}")
                    pipe.get(self._sender_domain_key(u, d))
            results = deque(pipe.execute())
            pipe.reset()
        return {
            d: {u: self._decode_policy_cache(results.popleft()) for u in users}
            for d in domains
        }



"""chapps.config

An easy way to encapsulate library-wide defaults and settings for CHAPPS
"""
import collections.abc
import configparser
from pathlib import Path
from os import environ as env
from chapps.util import AttrDict, VenvDetector
from chapps._version import __version__
import logging

try:
    from dotenv import load_dotenv

    load_dotenv()
except Exception:
    pass

logger = logging.getLogger(__name__)


class CHAPPSConfig:
    """Class wrapper around config logic makes testing easier"""

    # ultimately, we may need also to allow for a command-line option
    @staticmethod
    def what_config_file():
        """
        Determine what config file to read.  This is to allow for easier
           addition of a command-line option.
        """
        config_file = Path(env.get("CHAPPS_CONFIG", "/etc/chapps/chapps.ini"))
        return config_file

    @staticmethod
    def setup_config(cp):
        """Setup default config pattern on the parser passed in"""
        cp["CHAPPS"] = {
            "payload_encoding": "utf-8",
            "user_key": "sasl_username",
            "require_user_key": True,
            "no_user_key_response": "REJECT Rejected - Authentication failed",
            "password": (
                "effda33d276c1d5649f3933a6d6b286e"
                "d7eaaede0b944221e7699553ce0558e2"
            ),
        }
        cp["PolicyConfigAdapter"] = {
            "adapter": "mariadb",
            "db_host": "localhost",
            "db_port": "3306",
            "db_name": "chapps",
            "db_user": "chapps",
            "db_pass": "chapps",
        }
        cp["OutboundQuotaPolicy"] = {
            "listen_address": "localhost",
            "listen_port": 10225,
            "margin": 0.10,
            "min_delta": 0,
            "counting_recipients": True,
            "rejection_message": "REJECT Rejected - outbound quota fulfilled",
            "acceptance_message": "DUNNO",
            "null_sender_ok": False,
        }
        cp["GreylistingPolicy"] = {
            "listen_address": "localhost",
            "listen_port": 10226,
            "rejection_message": "DEFER_IF_PERMIT Service temporarily unavailable - greylisted",
            "acceptance_message": "DUNNO",
            "null_sender_ok": False,
        }
        cp["SPFEnforcementPolicy"] = {
            "listen_address": "localhost",
            "listen_port": 10227,
            "whitelist": [],
            "null_sender_ok": False,
        }
        cp["PostfixSPFActions"] = {
            "passing": "prepend",
            "fail": "550 5.7.1 SPF check failed: {reason}",
            "temperror": "451 4.4.3 SPF record(s) temporarily unavailable: {reason}",
            "permerror": "550 5.5.2 SPF record(s) are malformed: {reason}",
            "none_neutral": "greylist",
            "softfail": "greylist",
        }
        cp["SenderDomainAuthPolicy"] = {
            "listen_address": "localhost",
            "listen_port": 10225,
            "rejection_message": "REJECT Rejected - not allowed to send mail from this domain",
            "acceptance_message": "DUNNO",
            "null_sender_ok": False,
        }
        cp["Redis"] = {
            "sentinel_servers": "",
            "sentinel_dataset": "",
            "server": "localhost",
            "port": "6379",
        }
        return cp

    @staticmethod
    def write_config(cp, fn):
        """
        write config in cp to file named fn,
        if no such file exists; returns a Path
        object pointing at the config file
        """
        config_file = Path(fn)
        if not config_file.parent.exists():
            try:
                config_file.parent.mkdir(
                    0o777, True
                )  # attempt to make any missing parent directories
            except OSError as e:
                logger.error(
                    f"The specified config file's directory did not exist and could not be created.  File: {str(config_file)}"
                )
                raise e
        with config_file.open("w") as fh:
            cp.write(fh)
        return config_file

    def __init__(self):
        ### Create and initialize the config
        config_file = CHAPPSConfig.what_config_file()
        self.venvdetector = VenvDetector()
        self.configparser = configparser.ConfigParser(interpolation=None)
        CHAPPSConfig.setup_config(self.configparser)

        ### Initialize a config file if none
        if not config_file.exists():
            CHAPPSConfig.write_config(self.configparser, config_file)
        else:
            self.configparser.read(str(config_file))
        self.configparser["CHAPPS"]["config_file"] = str(config_file)
        self.configparser["CHAPPS"]["version"] = f"CHAPPS v{__version__}"
        self.configparser["CHAPPS"]["docpath"] = str(self.venvdetector.docpath)
        self.chapps = AttrDict(self.configparser["CHAPPS"])
        self.adapter = AttrDict(self.configparser["PolicyConfigAdapter"])
        self.actions_spf = AttrDict(self.configparser["PostfixSPFActions"])
        self.redis = AttrDict(self.configparser["Redis"])
        ### these are somewhat obsolete now
        self.policy_oqp = AttrDict(self.configparser["OutboundQuotaPolicy"])
        self.policy_sda = AttrDict(self.configparser["SenderDomainAuthPolicy"])
        self.policy_grl = AttrDict(self.configparser["GreylistingPolicy"])
        self.policy_spf = AttrDict(self.configparser["SPFEnforcementPolicy"])

    def get_block(self, blockname):
        try:
            return AttrDict(self.configparser[blockname])
        except Exception:
            pass
        return None

    def write(self, location=None):
        location = Path(location or self.chapps.config_file)
        config_file = self.chapps.config_file
        version = self.chapps.version
        docpath = self.chapps.docpath
        self.configparser.remove_option("CHAPPS", "config_file")
        self.configparser.remove_option("CHAPPS", "version")
        self.configparser.remove_option("CHAPPS", "docpath")
        result = CHAPPSConfig.write_config(self.configparser, location)
        if config_file:
            self.configparser["CHAPPS"]["config_file"] = config_file
        self.configparser["CHAPPS"]["version"] = version
        self.configparser["CHAPPS"]["docpath"] = docpath
        return result


config = CHAPPSConfig()



"""Common code between routers; mainly dependencies"""
from typing import Optional, List
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import IntegrityError
from fastapi import status, Depends, Body, HTTPException
from functools import wraps
import inspect
import logging
from chapps.rest.dbsession import sql_engine
from chapps.rest.models import AssocOperation, DeleteResp, TextResp
from chapps.rest.dbmodels import JoinAssoc
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)


async def list_query_params(
    skip: Optional[int] = 0,
    limit: Optional[int] = 1000,
    q: Optional[str] = "%",
):
    return dict(q=q, skip=skip, limit=limit)


def model_name(cls):
    return cls.__name__.lower()


def load_model_with_assoc(cls, assoc: List[JoinAssoc], engine=sql_engine):
    mname = model_name(cls)
    assoc_s = "_".join([a.assoc_name for a in assoc])
    fname = f"load_{mname}_with_{assoc_s}"
    Session = sessionmaker(engine)

    def get_model_and_assoc(item_id: int, name: Optional[str]):
        remarks = []
        items = {k: None for k in [mname, *[a.assoc_name for a in assoc]]}
        with Session() as sess:
            if item_id:
                items[mname] = sess.scalar(cls.select_by_id(item_id))
            if name and not items[mname]:
                items[mname] = sess.scalar(cls.select_by_name(name))
                if items[mname] and item_id:
                    remarks.append(
                        f"Selecting {mname} {items[mname].name} with "
                        f"id {items[mname].id} by name because "
                        "provided id does not exist."
                    )
            if not items[mname] and not (item_id or name):
                logger.debug(  # log this, as it is weird
                    f"{fname}({item_id!r}, {name!r}): unable to load {mname}"
                )
                raise HTTPException(  # describe error to the caller
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=(
                        f"One of {cls.id_name()} or name must be provided. "
                        "If both are provided, {cls.id_name()} is preferred."
                    ),
                )
            if items[mname]:
                for a in assoc:
                    items[a.assoc_name] = getattr(items[mname], a.assoc_name)
            else:
                detail = "No {mname} could be found with "
                if item_id:
                    detail += "id {item_id}"
                    if name:
                        detail += f" or with "
                if name:
                    detail += "name {name}"
                raise HTTPException(
                    status_code=status.HTTP_404_NOT_FOUND, detail=detail
                )
        return (items.pop(mname), items, remarks)

    get_model_and_assoc.__name__ = fname
    return get_model_and_assoc


def db_interaction(  # a decorator with parameters
    *,
    cls,
    engine=sql_engine,
    exception_message: str = ("{route_name}:{model}"),
    empty_set_message: str = ("Unable to find a matching {model}"),
):
    """
    the db_interaction decorator requires a couple of parameters,
    and provides optional arguments to override the messages for
    either of two eventualities:
    1. any exception occurs; the argument list is automatically appended
    2. the set of return values (from an access operation) is empty, OR
       a delete operation could not find any objects to delete
    """

    mname = model_name(cls)
    Session = sessionmaker(engine)

    def interaction_wrapper(rt_coro):
        exc = exception_message.format(
            route_name=rt_coro.__name__, model=mname
        )
        empty = empty_set_message.format(
            route_name=rt_coro.__name__, model=mname
        )

        @wraps(rt_coro)
        async def wrapped_interaction(*args, **kwargs):
            with Session() as session:
                rt_coro.__globals__["session"] = session
                rt_coro.__globals__["model_name"] = mname

                try:
                    result = await rt_coro(*args, **kwargs)
                    if result:
                        return result
                except Exception:
                    logger.exception(exc + f"({args!r},{kwargs!r})")
            raise HTTPException(status_code=404, detail=empty)

        wrapped_interaction.__doc__ = rt_coro.__doc__
        return wrapped_interaction  # a coroutine

    return interaction_wrapper  # a regular function


def get_item_by_id(cls, *, response_model, engine=sql_engine, assoc=None):
    """
    Build a route to get an item by ID:
    first argument is the main datamodel for the request
    named arguments supply the DB engine for session creation
    and the Pydantic response model for the output
    the optional dict assoc maps the names of associated models
    onto the data model for the associated objects
    """

    @db_interaction(cls=cls, engine=engine)
    async def get_i(item_id: int):
        stmt = cls.select_by_id(item_id)
        item = session.scalar(stmt)
        if item:
            if assoc:
                extra_args = {
                    key: model.wrap(getattr(item, key)) for model, key in assoc
                }
                return response_model.send(cls.wrap(item), **extra_args)
            else:
                return response_model.send(cls.wrap(item))

    get_i.__name__ = f"get_{model_name(cls)}"
    get_i.__doc__ = (
        f"Retrieve {cls.__name__} records by ID, "
        "along with all associated records."
    )
    return get_i


def list_items(cls, *, response_model, engine=sql_engine):
    """
    Build a route to list items.
    The factory just needs the control data -- the engine, the response model
    The returned closure expects to receive the query parameters as a dict,
    since that is what the dependency will yield.
    """

    @db_interaction(cls=cls, engine=engine)
    async def list_i(qparams: dict = Depends(list_query_params)):
        stmt = cls.windowed_list(**qparams)
        items = cls.wrap(session.scalars(stmt))
        if items:
            return response_model.send(items)

    list_i.__name__ = f"list_{model_name(cls)}"
    list_i.__doc__ = f"""
        List {cls.__name__}s.<br/>
        Pass a substring to match as `q`.<br/>
        Implement pagination by providing `skip` and `limit`.
        """
    return list_i


def list_associated(
    cls, *, assoc: JoinAssoc, response_model, engine=sql_engine
):
    """
    Build a route to list associated items with pagination
    """
    mname = model_name(cls)
    fname = f"{mname}_list_{assoc.assoc_name}"
    params = dict(item_id=int, qparams=dict)
    p_dfls = dict(item_id=None, qparams=Depends(list_query_params))

    @db_interaction(cls=cls, engine=engine)
    async def assoc_list(*pargs, **args):
        # item_id for source object of type cls
        # the specified association is listed according to qparams
        stmt = assoc.assoc_model.windowed_list_by_ids(
            subquery=assoc.select_ids_by_source_id(args["item_id"]),
            **args["qparams"],
        )
        assocs = assoc.assoc_model.wrap(session.scalars(stmt))
        if assocs:
            return response_model.send(assocs)

    routeparams = [
        inspect.Parameter(
            name=param,
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            annotation=type_,
            default=p_dfls[param],
        )
        for param, type_ in params.items()
    ]
    assoc_list.__signature__ = inspect.Signature(routeparams)
    assoc_list.__annotations__ = params
    assoc_list.__name__ = fname
    assoc_list.__doc__ = (
        f"List {assoc.assoc_name} objects associated with a particular {mname},"
        " identified by ID.<br/>Supply standard query parameters for"
        " pagination."
    )
    return assoc_list


def delete_item(
    cls, *, response_model=DeleteResp, params=None, engine=sql_engine
):
    f"""Delete {cls.__name__}"""
    params = params or dict(ids=List[int])
    mname = model_name(cls)

    @db_interaction(cls=cls, engine=engine)
    async def delete_i(item_ids: List[int]):
        try:
            session.execute(cls.remove_by_id(item_ids))
            session.commit()
        except IntegrityError:
            logger.exception("trying to delete {mname} with {args!r}")
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail="Database integrity conflict.",
            )
        return response_model.send()

    delete_i.__name__ = f"delete_{mname}"
    delete_i.__doc__ = f"""
        Accepts a list of {mname} object IDs in the body.
        Deletes them."""
    return delete_i


def adjust_associations(
    cls,
    *,
    assoc: List[JoinAssoc],
    assoc_op: AssocOperation,
    params: dict = None,
    response_model=TextResp,
    engine=sql_engine,
):
    """Build a route to add or subtract associations"""
    mname = model_name(cls)
    assoc_s = "_".join([a.assoc_name for a in assoc])
    fname = f"{mname}_{assoc_op}_{assoc_s}"
    params = params or dict(item_id=int)
    assoc_params = {a.assoc_id: a.assoc_type for a in assoc}

    @db_interaction(cls=cls, engine=engine)
    async def assoc_op_i(*pargs, **args):
        # item_id will be used for the source object, and assoc_ids will
        # be a list of associated ids to either remove or add associations
        # for, ignoring integrity errors arising from attempting to insert
        # duplicate associations; non-existent associations should not cause
        # errors when a query attempts to delete them.
        extras = {
            a.assoc_name: (a, args.pop(a.assoc_id))
            for a in assoc
            if a.assoc_id in args
        }
        item_id = args["item_id"]
        if extras:
            for assoc_name, (assc, vals) in extras.items():
                if not vals:
                    continue
                if assoc_op == AssocOperation.add:
                    stmt = assc.insert_assoc(item_id, vals)
                elif assoc_op == AssocOperation.replace:
                    stmt = assc.update_assoc(item_id, vals)
                else:
                    stmt = assc.delete_assoc(item_id, vals)
                try:
                    session.execute(stmt)
                except IntegrityError:
                    pass  # ignoring as stated above
            session.commit()
            return response_model.send("Updated.")
        return response_model.send("Empty request.")

    # this approach may seem laborious, but it supports multicolumn prikeys
    routeparams = [
        inspect.Parameter(
            name=param,
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            annotation=type_,
        )
        for param, type_ in params.items()
    ]
    routeparams.extend(
        [
            inspect.Parameter(
                name=param,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=type_,
            )
            for param, type_ in assoc_params.items()
        ]
    )
    assoc_op_i.__signature__ = inspect.Signature(routeparams)
    assoc_op_i.__annotations__ = params
    assoc_op_i.__name__ = fname
    assoc_op_i.__doc__ = f"""
        {str(assoc_op).capitalize()}
        {', '.join([a.assoc_name for a in assoc])} objects
        {'to' if assoc_op==AssocOperation.add else 'from'} {mname}<br/>
        Accepts mname id as `item_id`.
    """

    return assoc_op_i


def update_item(cls, *, response_model, assoc=None, engine=sql_engine):
    """
    Build a route to update items.
    """
    mname = model_name(cls)
    fname = f"update_{mname}"
    params = {mname: cls}  # we are updating objects of this type

    @db_interaction(cls=cls, engine=engine)
    async def update_i(*pargs, **args):
        extras = {}
        assoc_ret = {}
        if assoc:
            extras = {
                a.assoc_name: (a, args.pop(a.assoc_name))
                for a in assoc
                if a.assoc_name in args
            }
        item_id = args[mname].id
        stmt = cls.update_by_id(args[mname])
        try:
            result = session.execute(stmt)
            if result.rowcount > 0:  # commit if changes were made
                session.commit()
            else:  # return None if there is no record to update -> 404
                return
        except Exception:
            logger.exception(f"{fname}: {args!r}")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
            )
        item = session.scalar(cls.select_by_id(item_id))
        if extras:
            for assoc_name, (assc, vals) in extras.items():
                if not vals:
                    continue
                try:
                    session.execute(
                        assc.delete().where(
                            getattr(assc.table.c, assc.source_id) == item_id
                        )
                    )
                    session.execute(assc.insert(), assc.values(item, vals))
                    session.commit()
                    assoc_ret[assoc_name] = assc.assoc_model.wrap(
                        getattr(item, assoc_name)
                    )
                except IntegrityError:
                    logger.exception(
                        f"{fname}: associating {item} with"
                        f" {assoc_name}s {vals!r}"
                    )
                    raise HTTPException(
                        status_code=status.HTTP_409_CONFLICT,
                        detail=(
                            "Unable to create requested association to"
                            f"{assc.assoc_model.__name__.lower()} entries."
                            "  Please check object ids and try again."
                        ),
                    )
        return response_model.send(cls.wrap(item), **assoc_ret)

    routeparams = [
        inspect.Parameter(
            name=param,
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            default=Body(...),
            annotation=type_,
        )
        for param, type_ in params.items()
    ]
    if assoc:
        routeparams.extend(
            [
                inspect.Parameter(
                    name=a.assoc_name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=Body(None),
                    annotation=a.assoc_type,
                )
                for a in assoc
            ]
        )
    update_i.__signature__ = inspect.Signature(routeparams)
    update_i.__annotations__ = params
    update_i.__name__ = fname
    update_i.__doc__ = f"""
        Update {cls.__name__} by id.<br/>
        All {mname} attributes are required.<br/>
        Associations are not required, but if provided, will
        completely replace any existing association relationships
        of the same type.
        """
    return update_i


def create_item(
    cls, *, response_model, params=None, assoc=None, engine=sql_engine
):
    """
    Build a route to create items.
    """
    params = params or dict(name=str)
    fname = f"create_{model_name(cls)}"

    @db_interaction(cls=cls, engine=engine)
    async def create_i(*pargs, **args):
        """Bogus create item docstring"""
        extras = {}
        assoc_ret = {}
        if assoc:
            extras = {
                a.assoc_name: (a, args.pop(a.assoc_name))
                for a in assoc
                if a.assoc_name in args
            }
        item = cls.Meta.orm_model(**args)
        try:
            session.add(item)
            session.commit()
        except IntegrityError:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail="Unique key conflict.",
            )
        except Exception:
            logger.exception(f"{fname}: {args!r}")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
            )
        if extras:
            for assoc_name, (assc, vals) in extras.items():
                if not vals:  # no empty inserts
                    continue
                try:
                    session.execute(assc.insert(), assc.values(item, vals))
                    session.commit()
                    assoc_ret[assoc_name] = assc.assoc_model.wrap(
                        getattr(item, assoc_name)
                    )
                except IntegrityError:
                    raise HTTPException(
                        status_code=status.HTTP_409_CONFLICT,
                        detail=(
                            "Unable to create requested association to"
                            f"{assc.assoc_model.__name__.lower()} entries."
                            "  Please check object ids and try again."
                        ),
                    )

        return response_model.send(cls.wrap(item), **assoc_ret)

    routeparams = [  # assemble signature for FastAPI
        inspect.Parameter(
            name=param,
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            default=Body(...),
            annotation=type_,
        )
        for param, type_ in params.items()
    ]
    if assoc:
        routeparams.extend(
            [
                inspect.Parameter(
                    name=a.assoc_name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=Body(None),
                    annotation=a.assoc_type,
                )
                for a in assoc
            ]
        )
    create_i.__signature__ = inspect.Signature(routeparams)
    create_i.__annotations__ = params
    create_i.__name__ = fname
    create_i.__doc__ = f"""
        Create a new {cls.__name__} record in the database.<br/>
        All attributes are required.<br/>
        The new object will be returned, including its id.<br/>
        Raises descriptive errors on 409; checking the detail
          of the error may aid in debugging.
        """
    return create_i



"""CHAPPS module for controlling logging"""
import logging
from logging.handlers import SysLogHandler

DEFAULT_LEVEL = logging.DEBUG
DEFAULT_FACILITY = SysLogHandler.LOG_LOCAL0


class LogSetup:  # pragma: no cover
    debug_formatter = logging.Formatter(
        "CHAPPS:%(levelname)s %(filename)s@%(lineno)s: %(message)s"
    )
    maillog_formatter = logging.Formatter("CHAPPS:%(levelname)s %(message)s")

    syslog_handler = SysLogHandler(
        facility=DEFAULT_FACILITY, address="/dev/log"
    )
    syslog_handler.setLevel(DEFAULT_LEVEL)
    syslog_handler.setFormatter(maillog_formatter)  # or debug_formatter

    def __init__(self):
        if not logging.getLogger(None).hasHandlers():
            logging.basicConfig(handlers=[self.syslog_handler])


logsetup = LogSetup()
logger = logging.getLogger("chapps")
# without this, the library may not emit logs from a script
logger.setLevel(DEFAULT_LEVEL)



from fastapi import APIRouter, status
from chapps.rest.models import Quota, QuotaResp, QuotasResp, DeleteResp
from .common import (
    get_item_by_id,
    list_items,
    create_item,
    delete_item,
    update_item,
)
import logging
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)

api = APIRouter(
    prefix="/quotas",
    tags=["quotas"],
    responses={404: {"description": "Quota not found."}},
)


api.get("/", response_model=QuotasResp)(
    list_items(Quota, response_model=QuotasResp)
)


api.get("/{item_id}", response_model=QuotaResp)(
    get_item_by_id(Quota, response_model=QuotaResp)
)

api.post(
    "/",
    status_code=201,
    response_model=QuotaResp,
    responses={
        status.HTTP_400_BAD_REQUEST: {"description": "Unique key error."}
    },
)(
    create_item(
        Quota, response_model=QuotaResp, params=dict(name=str, quota=int)
    )
)

logger.debug("Created Quota::create_i")
# @api.post("/")
# async def create_quota(name: str=Body(...), limit: int=Body(...)):

api.delete("/", response_model=DeleteResp)(delete_item(Quota))

api.put("/", response_model=QuotaResp)(
    update_item(Quota, response_model=QuotaResp)
)



from typing import List
from fastapi import status, APIRouter
from ..models import (
    User,
    Quota,
    Domain,
    UserResp,
    UsersResp,
    DomainsResp,
    DeleteResp,
    IntResp,
    TextResp,
    AssocOperation,
)
from .common import (
    get_item_by_id,
    list_items,
    create_item,
    delete_item,
    update_item,
    adjust_associations,
    list_associated,
)
import logging
import chapps.logging

logger = logging.getLogger(__name__)

api = APIRouter(
    prefix="/users",
    tags=["users"],
    responses={404: {"description": "User not found."}},
)

user_quota_assoc = User.join_assoc(
    assoc_name="quota",
    assoc_type=int,
    assoc_model=Quota,
    assoc_id=Quota.id_name(),
    table=User.Meta.orm_model.metadata.tables["quota_user"],
)

user_domains_assoc = User.join_assoc(
    assoc_name="domains",
    assoc_type=List[int],
    assoc_model=Domain,
    assoc_id=Domain.id_name(),
    table=User.Meta.orm_model.metadata.tables["domain_user"],
)

user_join_assoc = [user_quota_assoc, user_domains_assoc]

api.post(
    "/",
    status_code=201,
    response_model=UserResp,
    responses={
        status.HTTP_400_BAD_REQUEST: {"description": "Could not create user."},
        status.HTTP_409_CONFLICT: {"description": "Unique key error."},
    },
)(create_item(User, response_model=UserResp, assoc=user_join_assoc))


api.delete("/", response_model=DeleteResp)(delete_item(User))


api.get("/", response_model=UsersResp)(
    list_items(User, response_model=UsersResp)
)


api.get("/{item_id}", response_model=UserResp)(
    get_item_by_id(
        User,
        response_model=UserResp,
        assoc=[(Quota, "quota"), (Domain, "domains")],
    )
)

api.get("/{item_id}/allowed/", response_model=DomainsResp)(
    list_associated(User, assoc=user_domains_assoc, response_model=DomainsResp)
)

api.put("/", response_model=UserResp)(
    update_item(User, response_model=UserResp, assoc=user_join_assoc)
)

api.put("/{item_id}/allow/", response_model=TextResp)(
    adjust_associations(
        User, assoc=[user_domains_assoc], assoc_op=AssocOperation.add
    )
)

api.put("/{item_id}/deny/", response_model=TextResp)(
    adjust_associations(
        User, assoc=[user_domains_assoc], assoc_op=AssocOperation.subtract
    )
)

api.put("/{item_id}/quota/{quota_id}")(
    adjust_associations(
        User, assoc=[user_quota_assoc], assoc_op=AssocOperation.replace
    )
)

# commenting out to get a clean release without these non-working routes
#
# we will provide these routes in a future release
# along with routes to count a user's domain authorizations
# and paginate the list of those authorizations


# @api.get("/count/", response_model=IntResp)
# async def count_all_users():
#     return await count_users("%")


# @api.get("/count/{pattern}", response_model=IntResp)
# async def count_users(pattern: str):
#     cur = pca.conn.cursor()
#     sanitized_pattern = pattern
#     query = f"SELECT COUNT( * ) FROM users WHERE name LIKE ?"
#     logger.debug(
#         f"Attempting to count users like {pattern} with query {query}"
#     )
#     cur.execute(query, (f"%{pattern}%",))
#     results = cur.fetchone()[0]
#     cur.close()
#     return IntResp.send(results)



from setuptools import setup
from setuptools.command.install import install
from pathlib import Path
import re
import sys
import atexit

import logging

# The service should have a role account to run as.  We recommend 'chapps'
SERVICE_ROLEACCOUNT = "chapps"
SERVICE_ROLEGROUP = SERVICE_ROLEACCOUNT
# This maps service-script locations in the repo to their descriptive names
SERVICE_DESCRIPTIONS = {
    "services/chapps_outbound_quota.py": "CHAPPS Outbound Quota Service",
    "services/chapps_outbound_multi.py": "CHAPPS Outbound Multipolicy Service",
    "services/chapps_greylisting.py": "CHAPPS Greylisting Service",
}
# create a list of service-script locations for the setup() call
SERVICES = SERVICE_DESCRIPTIONS.keys()

# set logging config; messages at INFO level for now
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


# Describe a class for doing some post-install work, mainly to set up
# SystemD service files for venvs
class PostInstallSetup(install):
    """
    This class can tell if a venv is in use,
    and format SystemD service startup files appropriately
    """

    # find the base prefix; hopefully pyenv-compatible
    def get_base_prefix_compat(self):
        return (
            getattr(sys, "base_prefix", None)
            or getattr(sys, "real_prefix", None)
            or sys.prefix
        )

    # return whether we are in a venv
    def in_virtualenv(self):
        return self.get_base_prefix_compat() != sys.prefix

    # build a closure to use with atexit.register()
    def post_install_setup_factory(self):
        """
        This function returns a closure suitable for performing
        the post-install setup
        """
        env_settings = {}
        service_descriptions = SERVICE_DESCRIPTIONS
        template_loc = Path(sys.prefix) / "chapps" / "install"
        service_home = Path(sys.prefix)
        service_bin = service_home / "bin"
        template_locations = [template_loc]
        if self.in_virtualenv():
            python_invocation = str(service_bin / "python3")
            gunicorn_invocation = str(service_bin / "gunicorn")
            (Path(sys.prefix) / "etc").mkdir(
                511, False, True
            )  # the mode is really a mask
            env_settings["CHAPPS_CONFIG"] = str(
                Path(sys.prefix) / "etc" / "chapps.ini"
            )
        else:
            python_invocation = "/usr/bin/env python3"
            gunicorn_invocation = "/usr/bin/env gunicorn"
            if (Path(sys.prefix) / "local").is_dir():
                possible_template_loc = (
                    Path(sys.prefix) / "local" / "chapps" / "install"
                )
                if possible_template_loc.is_dir():
                    template_locations.append(possible_template_loc)
                    service_bin = Path(sys.prefix) / "local" / "bin"
        template_locations.append(Path(sys.prefix).parent / "install")
        # define a routine for reading in the template
        # returns template contents followed by location Path object
        def template_reader(template_file):
            """This function finds and returns the contents of a template"""
            for template_dir in template_locations:
                template_path = template_dir / str(template_file)
                try:
                    with template_path.open() as tfh:
                        template = tfh.read()
                        return template, template_path
                except FileNotFoundError:
                    next
                except Exception:
                    logger.exception(f"reading {template_path}")
                    next
            return "", None

        # return the template filled out
        # using the namespace passed in as kwargs
        def fill_template(template, **kwargs):
            if len(env_settings) > 0 and "env_settings_lines" not in kwargs:
                kwargs["env_settings_lines"] = "\n".join(
                    [
                        f'Environment="{sym}={val}"'
                        for sym, val in env_settings.items()
                    ]
                )
            else:
                kwargs["env_settings_lines"] = ""
            return template.format(**kwargs)

        def post_install_setup():
            """
            This function performs the setup steps:
            template out service files
            """
            nonlocal python_invocation
            nonlocal gunicorn_invocation
            nonlocal service_home
            service_template, template_found = template_reader(
                "chapps-systemd-service.tmpl"
            )
            api_svc_template, api_svc_tmpl_path = template_reader(
                "chapps-api-systemd-service.tmpl"
            )
            services_to_enable = []
            service_roleaccount = SERVICE_ROLEACCOUNT
            service_rolegroup = SERVICE_ROLEGROUP
            if template_found:
                for (
                    service_source,
                    service_description,
                ) in service_descriptions.items():
                    # set up some variables used by the template
                    service_exec_path = service_bin / Path(service_source).name
                    service_file_path = template_found.parent / (
                        Path(service_source).stem + ".service"
                    )
                    try:
                        with service_file_path.open("w") as outfile:
                            outfile.write(
                                fill_template(service_template, **locals())
                            )
                    except Exception:
                        logger.exception(f"writing {service_file_path}")
                        next
                    else:  # runs if there are no exceptions
                        services_to_enable.append(service_file_path)
            if api_svc_tmpl_path:
                api_svc_path = (
                    api_svc_tmpl_path.parent / "chapps_rest_api.service"
                )
                try:
                    with api_svc_path.open("w") as outfile:
                        outfile.write(
                            fill_template(api_svc_template, **locals())
                        )
                except Exception:
                    logger.exception(f"writing {str(api_svc_path)}")
                else:
                    services_to_enable.append(api_svc_path)
            if len(services_to_enable) > 0:
                logger.warning(
                    "The following files were written to "
                    f"{template_found.parent}: "
                    + ", ".join([p.name for p in services_to_enable])
                )
            logger.warning(
                "No SystemD service file descriptions could be created."
            )

        return post_install_setup

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.template_found = [None]
        atexit.register(self.post_install_setup_factory())


VERSIONFILE = "chapps/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError(f"Unable to find version string in {VERSIONFILE}")
with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    cmdclass={"install": PostInstallSetup},
    name="chapps",
    packages=["chapps", "chapps.rest", "chapps.rest.routers"],
    data_files=[
        (
            "chapps",
            ["README.md", "INSTALLATION.md", "README-API.md", "CHANGELOG.md"],
        ),
        (
            "chapps/install",
            [
                "install/chapps-systemd-service.tmpl",
                "install/chapps-api-systemd-service.tmpl",
                "install/chapps-api-gunicorn.socket",
                "install/chapps-api-nginx.conf",
                "install/chapps-rsyslog.conf",
            ],
        ),
        (
            "chapps/postfix/greylisting",
            ["postfix/greylisting/main.cf", "postfix/greylisting/master.cf"],
        ),
        (
            "chapps/postfix/outbound_quota",
            [
                "postfix/outbound_quota/main.cf",
                "postfix/outbound_quota/master.cf",
            ],
        ),
        (
            "chapps/postfix/null_filter",
            ["postfix/null_filter/main.cf", "postfix/null_filter/master.cf"],
        ),
    ],
    scripts=[
        *SERVICES,
        "services/null_filter.py",
        "install/chapps_database_init.py",
    ],
    version=verstr,
    license="MIT",
    description="Caching, Highly-Available Postfix Policy Service",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Caleb S. Cullen",
    author_email="ccullen@easydns.com",
    url="https://gitlab.int.easydns.net/ccullen/chapps",
    download_url=f"https://github.com/easydns/chapps/tree/main/dist/chapps-{verstr}.tar.gz",
    keywords=["Postfix", "Policy", "Daemon"],
)



#!/usr/bin/env python3
"""Caching, Highly-Available Postfix Policy Service - Greylisting"""

### requires the python-pidfile library from https://github.com/mosquito/python-pidfile
import asyncio, pidfile, signal, functools
from chapps.switchboard import GreylistingHandler
from chapps.config import config
from chapps.signals import SignalHandlerFactory
import logging, chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def install_asyncio_signal_handlers(loop):
    for signame in {"SIGTERM", "SIGINT"}:
        sig = getattr(signal, signame)
        loop.add_signal_handler(
            sig, functools.partial(SignalHandlerFactory.signal_handler(loop), sig)
        )


async def main():
    """The grand shebang"""
    logger.debug("Starting CHAPPS Greylisting...")
    try:
        with pidfile.PIDFile("/tmp/chapps_greylisting.pid"):
            logger.debug("CHAPPS Greylisting service started.")
            handler = GreylistingHandler()
            handle_policy_request = handler.async_policy_handler()
            install_asyncio_signal_handlers(asyncio.get_running_loop())
            srv = await asyncio.start_server(
                handle_policy_request,
                handler.listen_address,
                handler.listen_port,
                start_serving=False,
            )
            await srv.serve_forever()
    except pidfile.AlreadyRunningError:
        logger.exception("CHAPPS Greylisting is already running. Exiting.")
    except asyncio.exceptions.CancelledError:
        logger.debug("CHAPPS Greylisting exiting on signal.")


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except Exception:
        logger.exception("UNEX")



"""SQLAlchemy DB session global setup"""
from sqlalchemy import create_engine
from sqlalchemy.engine import URL
from chapps.config import config

DIALECT_MAP = dict(
    mariadb="mysql",
    mysql="mysql",
)

def create_db_url(cfg=None):
    cfg = cfg or config
    adapter = cfg.adapter
    if adapter.adapter not in DIALECT_MAP:
        raise ValueError(
            (
                "Configured database adapter must be one of:"
                f"{', '.join([repr(v) for v in DIALECT_MAP.keys()])}"
            )
        )
    dialect = DIALECT_MAP[adapter.adapter]
    creds = dict(
        password=adapter.db_pass,           # auto encoded
        username=adapter.db_user,
        host=adapter.db_host or '127.0.0.1',
        port=adapter.db_port or '3306',
        database=adapter.db_name,
    )
    return URL.create(dialect, **creds)


sql_engine = create_engine(create_db_url())



"""chapps.util

These are the utility classes for CHAPPS
"""
from collections.abc import Mapping
import re
import logging
import sys
from pathlib import Path

logger = logging.getLogger(__name__)


class VenvDetector:
    def __init__(self, *, datapath=None):
        if datapath:
            self.datapath = Path(datapath)
        elif self.ve:
            self.datapath = Path(sys.prefix)
        else:
            self.datapath = Path("/usr/local")

    # find the base prefix; hopefully pyenv-compatible
    def get_base_prefix_compat(self):
        return (
            getattr(sys, "base_prefix", None)
            or getattr(sys, "real_prefix", None)
            or sys.prefix
        )

    # return whether we are in a venv
    def in_virtualenv(self):
        return self.get_base_prefix_compat() != sys.prefix

    @property
    def ve(self):
        if "_ve" not in vars(self):
            self._ve = self.in_virtualenv()
        return self._ve

    @property
    def docpath(self):
        if "_docpath" not in vars(self):
            self._docpath = self.datapath / "chapps"
        return self._docpath


class AttrDict:
    """
    This simple class allows accessing the keys of a hash as attributes on
    an object.  As a useful side effect it also casts floats and integers
    in advance.  This object is used for holding the configuration data.
    """

    boolean_pattern = re.compile("^[Tt]rue|[Ff]alse$")

    def __init__(self, data=None, **kwargs):
        if not data:
            data = kwargs
        for k, v in data.items():
            if k[0:2] != "__":
                val = v
                try:
                    val = int(v)
                except ValueError:
                    try:
                        val = float(v)
                    except ValueError:
                        m = self.boolean_pattern.match(v)
                        if m:
                            val = (
                                m.span(0)[1] == 4
                            )  # if the match is 4 chars long, it is True
                except TypeError:
                    pass
                setattr(self, k, val)


class PostfixPolicyRequest(Mapping):
    """
    An implementation of Mapping which by default only processes and caches
    values from the data payload when they are accessed, to avoid a bunch of
    useless parsing.
    """

    def __init__(self, payload):
        """
        We get passed a payload of strings
        which are formatted as 'key=val'
        """
        self._payload = payload[0:-2]

    # the main reason for this class:
    # find and memoize as attributes the values in the request payload
    # means we cannot test existence of attrs by getattr or self.<attr>
    def __getattr__(self, attr, dfl=None):
        """Overload in order to search for missing attributes in the payload"""
        if attr[0] == "_":  # leading underscores do not occur in the payload
            return None
        line = next(
            (l for l in self._payload if attr == l.split("=")[0]), None
        )
        if line:
            key, value = line.split("=")
            setattr(self, key, value)
            return value
        else:
            logger.debug(f"No lines in {self} matched {attr}.")
        return None

    # Since the datastructure can function as a hash, provide optimization
    def __getitem__(self, key):
        """
        Getting an item should optimize the result
        via the attribute mechanism
        """
        return getattr(self, key)

    # The datastructure is iterable, in case we need to enumerate the request
    def __iter__(self):
        """Return an iterable representing the mapping"""
        if not getattr(self, "_mapping", None):
            self._mapping = {
                k: v for k, v in [l.split("=") for l in self._payload]
            }
            # Since we end up parsing the entire payload
            # optimize it for future random access
            for k, v in self._mapping.items():
                setattr(self, k, v)
        yield from self._mapping

    # The length of the PPR is considered to be the number of items stored
    def __len__(self):
        """Act like a dict and return the number of k,v pairs"""
        return len(self._payload)

    # Representations of PPR use their own class name, and otherwise dump
    # the payload, but not a list of attributes
    def __repr__(self):
        """Dump the data in _payload"""
        return "%s( %r )" % (self.__class__.__name__, self._payload)

    # In order to use memoization with PPRs, a hash function is required
    def __hash__(self):
        """Create a reliable hash for this PPR"""
        if "_hash" not in vars(self):
            self._hash = hash(f"{self.instance}:{self.queue_id}")
        return self._hash

    # This convenience property provides a list of recipient email addresses
    # since the line may contain more than one
    @property
    def recipients(self):
        """
        A convenience method to split the 'recipient' datum
        into comma-separated tokens, for easier counting.
        """
        if "_recipients" not in vars(self):
            self._recipients = (
                self.recipient.split(",")
                if self.recipient and len(self.recipient) > 0
                else []
            )
        return self._recipients



"""chapps.switchboard

Message multiplexing objects and/or routines for CHAPPS
"""
from chapps.config import config  # the global instance of the config object
from chapps.policy import OutboundQuotaPolicy, GreylistingPolicy, SenderDomainAuthPolicy
from chapps.spf_policy import SPFEnforcementPolicy
from chapps.util import AttrDict, PostfixPolicyRequest
from chapps.outbound import OutboundPPR
from chapps.signals import (
    CHAPPSException,
    CallableExhausted,
    NullSenderException,
    AuthenticationFailureException,
)
from functools import cached_property
import logging
import asyncio

logger = logging.getLogger(__name__)  # pragma: no cover


class RequestHandler:
    """Wrap handling in an object-oriented factory so we can supply a policy"""

    def __init__(self, policy, *, pprclass=PostfixPolicyRequest):
        self.policy = policy
        self.config = self.policy.config  # in case a custom config is in use
        self.pprclass = pprclass

    @cached_property
    def listen_address(self):
        return self.policy.params.listen_address

    @cached_property
    def listen_port(self):
        return self.policy.params.listen_port

    def async_policy_handler(self):
        """Returns a coroutine which handles requests by to the policy"""
        pprclass = self.pprclass
        policy = self.policy
        accept = policy.params.acceptance_message
        reject = policy.params.rejection_message
        encoding = config.chapps.payload_encoding
        logger.debug(
            f"Policy handler requested for {type(policy).__name__}"
            f" using PPR class {pprclass.__name__}."
        )

        async def handle_policy_request(reader, writer) -> None:
            """Handles reading and writing the streams around policy messages"""
            while True:
                try:
                    policy_payload = await reader.readuntil(b"\n\n")
                except ConnectionResetError:
                    logger.debug("Postfix said goodbye. Terminating this thread.")
                    return
                except asyncio.IncompleteReadError as e:
                    logger.debug(
                        "Postfix hung up before a read could be completed. Terminating this thread."
                    )
                    return
                except CallableExhausted as e:
                    raise e
                except Exception:
                    logger.exception("UNEXPECTED ")
                    if reader.at_eof():
                        logger.debug(
                            "Postfix said goodbye oddly. Terminating this thread."
                        )
                        return
                    continue
                logger.debug(f"Payload received: {policy_payload.decode('utf-8')}")
                policy_data = pprclass(policy_payload.decode(encoding).split("\n"))
                if policy.approve_policy_request(policy_data):
                    resp = ("action=" + accept + "\n\n").encode()
                    logger.debug(f"  .. Accepted.  Sending {resp}")
                else:
                    resp = ("action=" + reject + "\n\n").encode()
                    logger.debug(f"  .. Rejected with {resp}")
                try:
                    writer.write(resp)
                except asyncio.CancelledError:  # pragma: no cover
                    pass
                except Exception:
                    logger.exception(f"Exception raised trying to send {resp}")
                    return

        return handle_policy_request


class CascadingPolicyHandler:
    """A handler class which cascades multiple yes/no policies"""

    def __init__(self, policies=[], *, pprclass=PostfixPolicyRequest):
        self.policies = policies
        self.pprclass = pprclass
        if not self.policies:
            raise ValueError("A list of policy objects must be provided.")
        self.config = self.policies[0].config  # all copies of the config are the same

    @cached_property
    def listen_address(self):
        return next(
            (getattr(p.params, "listen_address", None) for p in self.policies), None
        )

    @cached_property
    def listen_port(self):
        return next(
            (getattr(p.params, "listen_port", None) for p in self.policies), None
        )

    ### an asynchronous policy handler which cascades through all the policies; fails stop execution
    def async_policy_handler(self):
        """Returns a coroutine which handles requests according to the policies, in order"""
        pprclass = self.pprclass
        policies = self.policies
        encoding = self.config.chapps.payload_encoding
        logger.debug(
            f"Cascading policy handler requested for "
            f"{[ type(p) for p in policies ]} using PPR "
            f"class {pprclass.__name__}."
        )

        async def handle_policy_request(reader, writer):
            """Handles reading and writing the streams around policy approval messages, and manages the cascade"""
            while True:
                try:
                    policy_payload = await reader.readuntil(b"\n\n")
                except ConnectionResetError:
                    logger.debug("Postfix said goodbye. Terminating this thread.")
                    return
                except asyncio.IncompleteReadError as e:
                    logger.debug(
                        "Postfix hung up before a read could be completed. Terminating this thread."
                    )
                    return
                except CallableExhausted as e:
                    raise e
                except Exception:
                    if reader.at_eof():
                        logger.debug(
                            "Postfix said goodbye oddly. Terminating this thread."
                        )
                        return
                    else:
                        logger.exception("UNEXPECTED ")
                    continue
                logger.debug(f"Payload received: {policy_payload.decode( 'utf-8' )}")
                policy_data = pprclass(policy_payload.decode(encoding).split("\n"))
                approval = True
                for policy in policies:
                    try:
                        if policy.approve_policy_request(policy_data):
                            resp = "action=" + policy.params.acceptance_message + "\n\n"
                            logger.info(f"{type(policy).__name__} PASS {policy_data}")
                        else:
                            resp = "action=" + policy.params.rejection_message + "\n\n"
                            approval = False
                            logger.info(f"{type(policy).__name__} FAIL {policy_data}")
                    except NullSenderException:
                        if policy.params.null_sender_ok:
                            resp = "action=" + policy.params.acceptance_message + "\n\n"
                            logger.info(f"{type(policy).__name__} PASS NS {policy_data}")
                        else:
                            resp = "action=" + policy.params.rejection_message + "\n\n"
                            approval = False
                            logger.info(f"{type(policy).__name__} FAIL NS {policy_data}")
                    except AuthenticationFailureException:
                        resp = "action=" + config.no_user_key_response + "\n\n"
                        approval = False
                        logger.info(f"{type(policy).__name__} FAIL NA {policy_data}")
                    except CHAPPSException:
                        logger.exception("During policy evaluation:")
                    if not approval:
                        break
                try:
                    writer.write(resp.encode())
                except asyncio.CancelledError:
                    pass
                except Exception:
                    logger.exception(f"Exception raised trying to send {resp.strip()}")
                    return
        return handle_policy_request


class OutboundMultipolicyHandler(CascadingPolicyHandler):
    """Could be thought of as a concrete subclass of CPH, but meant more as a convenience"""

    def __init__(
        self, policies=[], *, pprclass=OutboundPPR
    ):  # note that we default to OutboundPPR here
        policies = policies or [
            SenderDomainAuthPolicy(),
            OutboundQuotaPolicy(),
        ]  # create a list of relevant outbound y/n policies
        super().__init__(policies, pprclass=pprclass)


class OutboundQuotaHandler(RequestHandler):
    def __init__(self, policy=None):
        p = policy or OutboundQuotaPolicy()
        super().__init__(p, pprclass=OutboundPPR)


class GreylistingHandler(RequestHandler):
    def __init__(self, policy=None):
        p = policy or GreylistingPolicy()
        super().__init__(p)


class SenderDomainAuthHandler(RequestHandler):
    def __init__(self, policy=None):
        p = policy or SenderDomainAuthPolicy()
        super().__init__(p, pprclass=OutboundPPR)


class SPFEnforcementHandler(RequestHandler):
    def __init__(self, policy=None):
        p = policy or SPFEnforcementPolicy()
        super().__init__(p)

    def async_policy_handler(self):
        """Returns a coroutine which handles requests according to the policy"""
        ### This override version for SPF enforcement does not assume a yes-or-no response pattern
        logger.debug(f"Policy handler requested for {type(self.policy).__name__}.")
        policy = self.policy
        encoding = config.chapps.payload_encoding

        async def handle_policy_request(reader, writer):
            """Handles reading and writing the streams around policy approval messages"""
            while True:
                try:
                    policy_payload = await reader.readuntil(b"\n\n")
                except ConnectionResetError:
                    logger.debug("Postfix said goodbye. Terminating this thread.")
                    return
                except CallableExhausted as e:
                    raise e
                except Exception:
                    logger.exception("UNEXPECTED ")
                    if reader.at_eof():
                        logger.debug(
                            "Postfix said goodbye oddly. Terminating this thread."
                        )
                        return
                    continue
                logger.debug(f"Payload received: {policy_payload.decode(encoding)}")
                policy_data = PostfixPolicyRequest(
                    policy_payload.decode(encoding).split("\n")
                )
                action = policy.approve_policy_request(policy_data)
                resp = ("action=" + action + "\n\n").encode()
                logger.debug(f"  .. SPF Enforcement sending {resp}")
                try:
                    writer.write(resp)
                except asyncio.CancelledError:  # pragma: no cover
                    pass
                except Exception:
                    logger.exception(f"Exception raised trying to send {resp}")
                    return

        return handle_policy_request
